/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
// assure non-reentrance
#ifndef MYGLOBAL_H
#define MYGLOBAL_H
#include <SDL/SDL.h>

#define SDLFLAGS SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_FULLSCREEN
#define iabs(x) (((x)<0)?-(x):(x))
#define sqr(x) ((x)*(x))
#define clamp(a) (a<0?0:(a>255?255:(a)))

//Pithagor rullz
#define pith0(x1, y1, z1, x2, y2, z2) sqrt(sqr(x1-x2)+sqr(y1-y2)+sqr(z1-z2))
#define pith(a, b) sqrt(sqr(a[0]-b[0])+sqr(a[1]-b[1])+sqr(a[2]-b[2]))
//#define inrange(x, y) ((unsigned)(x)<640 && (unsigned)(y)<480)


// this will toggle between flipping and just drawing on the framebuffer
#define ACTUALLYDISPLAY

// this will select whether CPU speed should be shown in the result screen
#define SHOW_CPU_SPEED

#define FONT_XSIZE 10.66666666666
#define FONT_YSIZE 18
#define FONT_XSIZE_FLOOR 10
#define FONT_XSIZE_CEIL (FONT_XSIZE_FLOOR+1)

#define RES_MAXX 1280
#define RES_MAXY 960
#define MAX_SPHERES 1300
#define MAX_SPHERE_SIDES 128
// tune the following constant up if the spheres look too choppy to you.
#define MIN_SPHERE_SIDES 12
#define EPS 0.0000000001
#define MAX_RAYTRACE_ITERATIONS 9
#define RAYTRACE_BLEND_COLOR 0x000000
#define BRIGHTNESS_C 25.0

// if a ray is dropped to the floor engine it checks if the |y| component of the vector is lower than this threshold.
// The idea is that such rays go in the infinity and should be painted black. The texturer should paint them anyway, but due to
// some nasty inaccuracy in the fixedpoint math some full intensity white points are introduced, garbling the picture.
#define DST_THRESHOLD 0.0055


#define FOV_CORRECTION 0.882
// this defines the Field-Of-View constant. 1 means 90 degrees.
//90 degrees is usually best (determined experimentally). More than 110 degrees are *FREAKY*
#define STANDARD_FIELD_OF_VIEW 1

//#define BLUR 1

// enable this to create a file with the coordinates for each frame generated
//#define RECORD 1
#define RECORD_FILE "rekord.txt"

// if the following define exists then no time calculations are performed: frames are rendered
// consecutively from the data given in the context file.
#define IGNORE_TIME 1


// this would create a checksum for all frames, if uncommented
//#define MAKE_CHECKSUM 1

// these are the phong shading coefficients. Change them as you like, but try to keep their sum below 1.
#define ambient 0.2
#define diffuse 0.55
#define specular 0.15

// These are all used in the sphere flags. They can be ORed to be in effect simultaneously

// the sphere will be raytraced to a single reflection
#define RAYTRACED	0x00000001

// allows multiple reflections
#define RECURSIVE	0x00000002

// the sphere will be traced through (will be transparent)
#define TRANSPARENT	0x00000004

// the sphere will never bounce, have gravity, be animated .. etc.
#define STATIC		0x00000010
// object can collide
#define COLLIDEABLE	0x00000020
// new coords are calculated each frame using a corresponding AniSphere structure
#define ANIMATED	0x00000040
// object will be affected by the gravity
#define GRAVITY		0x00000100
// object will be affected by the air resistance
#define AIR		0x00000200

// object will be mapped with the given texture
#define MAPPED		0x00001000


/****************************************************
 * These defines did belong to fract.cpp            *
 ****************************************************/
#define MAX_DIST 1024*1024
#define MAX_TEXTURE_SIZE	256
#define LOG2_MAX_TEXTURE_SIZE	8
#define LOG2_TEXTURE_AREA	(LOG2_MAX_TEXTURE_SIZE*2)
#define TEX_S 3
#define BALL_T_AND_MASK ((MAX_TEXTURE_SIZE>>TEXTURE_FOR_BALLS)-1)
#define r(x) ((x+32768)/65536)
#define i2im(a, b) ((((unsigned) a) * ((unsigned) b))>>16)
#define PDIVIZOR 8
// the log2(PDIVIZOR):
#define DIVIZOR_SHIFT 3

#define __512 256*1024

// the `8' here should be LOG2_MAX_TEXTURE_SIZE
#define SHIFT_LEFT_LOG2_OF_MAX_TEXTURE_SIZE_mm2 "	pslld	$8,	%%mm2\n"
#define SHIFT_LEFT_LOG2_OF_MAX_TEXTURE_SIZE_mm3 "	pslld	$8,	%%mm3\n"

#define SHOWFPS
//#define FSTAT
#define BILINEAR_FILTER

#ifdef SHOWFPS
// update the FPS thing two times per second
#define FPS_UPDATE_INTERVAL 500
#endif

const char flag2str[32][16] = {
"raytraced", "recursive", "transparent", "", "static", "collideable", "animated", "", "gravity", "air", "", "", "mapped", "", "", "",
"","","","","","","","","","","","","","","",""};

typedef struct {
int x, y;
int *data;
}RawImg;

typedef struct sphere{
double x, y, z, d; // coordinates
double xi, yi, zi; // movement vector
double mass, time; /* Here, `time' is expressed in seconds and means that the coordinates
			of the sphere are valid at that time only. This allows to have spheres
			that are different in age. At the end of each frame calculation, `time'-s of
			all spheres are equalized and the needed movement is processed by then */
double dist; //distance from the user
float refl; //reflection strength
float opacity; // opacity if the object is TRANSPARENT
int r, g, b; //sphere color
int flags  ;
int AniIndex;// index in the array (see sphere.cpp). Used if the object is ANIMATED
RawImg *tex; // used if the object is MAPPED - points to the texture to map.
}; // sizeof(sphere) == 108

#include "bitmap.h"
#include "gfx.h"
#include "sphere.h"
#include "vectormath.h"
#include "rgb2yuv.h"
#include "cmdline.h"
#include "saveload.h"
#include "fract.h"
#include "render.h"
#include "antialias.h"
#include "shaders.h"
#include "profile.h"

#endif //MYGLOBAL_H
