#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "MyGlobal.h"

using namespace std;

const char *floor_texture = "data/2.bmp";
//const char *floor_texture = "../../data/fract-texture.bmp";
const char *default_font = "data/font1.bmp";
const char *introtex = "data/intro.bmp";


SDL_Surface *screen;
SDL_Overlay *ovr=NULL;
RawImg tex;

// the default font

RawImg font0;

// precalculated storage...

int *sqrtsqrt;

// the framebuffer:

Uint32 framebuffer[RES_MAXX*RES_MAXY];
#ifdef BLUR
Uint32 blurbuffer[RES_MAXX*RES_MAXY], bb1[RES_MAXX*RES_MAXY];
#endif

// texture collection:

RawImg T[12]; // this allows texture sizes up to 4096 ;)
int end_tex;

// sphere collection

sphere sp[MAX_SPHERES];
double pp[MAX_SPHERES][MAX_SPHERE_SIDES][3];
int num_sides[MAX_SPHERES];
int spherecount;


int vframe=0, bilfilter=0;

Uint32 clk;
double daFloor = 0.0, daCeiling = 200.0;
int sse_enabled=0, mmx2_enabled=0, mmx_enabled=0;
double gX, gY;
double alpha = -0.95, beta = -0.15;
double w[3][3], cur[3]={30, 45, 50};
double pdlo[MAX_SPHERES];
double lft, delta;
double speed = 25;
double anglespeed = 0.2;
double mouse_sensitivity = 0.16, fov = STANDARD_FIELD_OF_VIEW;
double keyboard_sensitivity = 3.0;

int not_fps_written_yet = 1;
int runmode = 0;
int camera_moved;

int loops_remaining=9999;
int developer = 0;

// the light source
int lx = 200, ly=100, lz = 500;
double light_angle = 0;

extern int CollDetect;
extern int Physics;
extern double play_time;
extern int cd_frames;
extern double Area_const;
int kbd_control = 1;
int file_control = 0;
int used_threads = 0;
int use_shader = 0;
#ifdef MAKE_CHECKSUM
Uint32 cksum=0;
#endif

#ifdef SHOWFPS
Uint32 fpsclk;
int fpsfrm=0;
#endif

SDL_Surface *get_screen(void) {return screen;}
double bTime(void)
{
 return ((SDL_GetTicks() - clk) / 1000.0);
}

int lform( int x, int y, int lx, int ly)
{int z=(((int) (sqr(x-lx)+sqr(y-ly)))>>DIVIZOR_SHIFT);
 if ((unsigned) z<MAX_DIST)
 	{return sqrtsqrt[z];}
	else
	{return (int) (lightformulae_tiny(z<<DIVIZOR_SHIFT)*65536.0);}
}

void precalc(void)
{int i, sz;
 CPUCaps x;

 GetCPUCaps(&x);

  // the integrated multiplycolor + lightformulae assembly code requires SSE & MMX2
 // I don't know of a processor which supports SSE, but does NOT MMX2; nevermind I make MMX2 required here, just in case...
 if (!OptionExists("--no-sse"))
 	sse_enabled = x.hassse && x.hasmmx2;

// the assembly bilinear filtering routine requires MMX and MMX extended instruction sets.
 if (!OptionExists("--no-mmx2"))
 	mmx2_enabled = x.hasmmx2;

// the fast bluring routine requires only MMX
 if (!OptionExists("--no-mmx"))
 	mmx_enabled = x.hasmmx;

 // if we don't have SSE, we will precalculate the distance values. This will get us close to the calculation performance of using SSE,
 // but the integrated version will still be faster because of its integration.
 if ( !sse_enabled ) {
  	if ((sqrtsqrt = (int *) malloc(MAX_DIST*sizeof(int))) == NULL) {
 		printf("Cannot get memory for precalculated sqrtsqrt array!\n");
		exit(1);
 		}
 	// precalculate distance array (double sqrt calculations)
 	for (i=0;i<MAX_DIST*PDIVIZOR;i++) {
		sqrtsqrt[i/PDIVIZOR] = (int) (lightformulae_tiny(i)*65536.0);
		if ((i&(__512-1))==0)
			intro_progress(screen, 0.9 * (double)i/ (double)  (MAX_DIST*PDIVIZOR));
		}
	} else sqrtsqrt = NULL;
 // precalculate smaller (and possibly larger) textures
 T[0] = tex;
 sz = MAX_TEXTURE_SIZE/2;
 i=0;
 while (sz) {
 	i++;
	Shrink(T[i-1], T+i);
	sz/=2;
	intro_progress(screen, 0.9 + 0.1*i/log2(MAX_TEXTURE_SIZE));
 	}
 end_tex = i;
 gfx_init();

 cur[0]=666;
 cur[1]=90;
 cur[2]=250;

 alpha = 3*M_PI_2+0.21;
 beta = 0+0.0001;

 SDL_ShowCursor(0);
 camera_moved = 1;
}

void init_fract_array(void)
{int i;
 double btm;
 btm = bTime();
 create_fract_array(sp, &spherecount);
 btm = bTime();
 for (i=0;i<spherecount;i++) {
 	sp[i].time = btm;
	}
}

void freetextures(void)
{int i=0, sz = MAX_TEXTURE_SIZE;
 while (sz) {
 	FreeRaw(T+i);
	i++; sz/=2;
 	}
}

void CloseProgram(void)
{
 if (sqrtsqrt != NULL)
 	free(sqrtsqrt);
 gfx_close();
 bitmap_close();
 cmdline_close();
 rgb2yuv_close();
 sphere_close(sp, spherecount);
 vectormath_close();
 saveload_close();
 shaders_close();
}

void hwaccel_init(void)
{
 SDL_VideoInfo *vi;
 static char vdname[256];
 vi = (SDL_VideoInfo *) SDL_GetVideoInfo();
 SDL_VideoDriverName(vdname, 256);
#ifdef SHOW_VIDEO_INFO
 printf("Video Info: ");
 printf("-- video driver: [%s]\n", vdname);
 printf("->flags: 0x%X\n", screen->flags);
 printf("->hw_available: %d\n->blit_hw: %d\n->blit_hw_CC: %d\n->blit_hw_A: %d\n->blit_sw: %d\n->blit_sw_CC: %d\n->blit_sw_A: %d\n",
         vi->hw_available, vi->blit_hw, vi->blit_hw_CC, vi->blit_hw_A, vi->blit_sw, vi->blit_sw_CC, vi->blit_sw_A);
#endif
 if (!vi->hw_available)
 	printf("No hardware acceleration for surface-to-screen blits.\n");
 if (OptionExists("--no-overlay")) return;
 if (!vi->hw_available || OptionExists("--force-overlay")) {
 	printf("Will try to use YUV overlay%s...",(vi->hw_available?"":" to speed things up"));
 	ovr = SDL_CreateYUVOverlay(xres(), yres(), SDL_YUY2_OVERLAY, screen);
	if (ovr == NULL) {printf("FAILED\nCan't create YUY2 overlay!"); if (OptionExists("--force-overlay")) printf(" Won't use it no matter how much you --force it."); printf("\n"); }
	else {
		printf("OK\n");
 		if (ovr->hw_overlay || OptionExists("--force-overlay")) {
 			if (ovr->hw_overlay)
				printf("Overlay is hardware accelerated :-)\n");
			if (!OptionExists("--force-overlay"))
				printf("To disable overlay usage, add --no-overlay to the command line.\n");
			init_yuv_convert(0);
			}
			else {
			printf("Overlay is not hardware accelerated... falling back to software surface flipping (possibly too slow!)\nHint: Check your X server configuration, try using full hardware acceleration!\n");
			printf("If you want to force using overlay and you know what you're doing, you can accomplish it by passing\nthe --force-overlay option to the program.\n");
			SDL_FreeYUVOverlay(ovr);
			ovr = NULL;
			}
		}
	}
 printf("\n");
}

void kbd_tiny_do(int *ShouldExit)
{
	SDL_Event e;

	while ( SDL_PollEvent(&e)) { // handle keyboard & mouse
		if ( e.type == SDL_QUIT ) *ShouldExit = 1;
		if ( e.type == SDL_KEYDOWN ) {
			if (e.key.keysym.sym == SDLK_ESCAPE) *ShouldExit = 1;
			if (e.key.keysym.sym == SDLK_F12) { // F12 is screenshot shortcut
				RawImg tempf;
				tempf.data = NULL;
				CreateRaw(&tempf, xres(), yres());
				memcpy(tempf.data, get_image_buffer(), xres()*yres()*4);
				Take_Snapshot(tempf);
				FreeRaw(&tempf);
				}
			if (developer) {
				if (e.key.keysym.sym == SDLK_F11) { // F11 is bilinear filtering toggle shortcut
					bilfilter = !bilfilter;
					}
				if (e.key.keysym.sym == SDLK_F8) {
					switch (fsaa_mode) {
						case FSAA_MODE_NONE:    set_fsaa_mode (FSAA_MODE_4X_LO_Q); break;
						case FSAA_MODE_4X_LO_Q: set_fsaa_mode (FSAA_MODE_4X_HI_Q); break;
						case FSAA_MODE_4X_HI_Q: set_fsaa_mode (FSAA_MODE_NONE); break;
					}
				}
				if (e.key.keysym.sym == SDLK_F9 ) fov/=1.01;
				if (e.key.keysym.sym == SDLK_F10) fov*=1.01;
// THESE WERE USED TO ADJUST THE AREA CONSTANT
//				if (e.key.keysym.sym == SDLK_F3 ) Area_const *= 0.9;
//				if (e.key.keysym.sym == SDLK_F4 ) Area_const *= 1.1;
			}
		}
	}
}

void kbd_do(int *ShouldExit)
{
	SDL_Event e;
	Uint8 *keystate;
	int deltax, deltay;

	while ( SDL_PollEvent(&e)) { // handle keyboard & mouse
		if ( e.type == SDL_QUIT ) *ShouldExit = 1;
		if ( e.type == SDL_KEYDOWN ) {
			if (e.key.keysym.sym == SDLK_ESCAPE) *ShouldExit = 1;
			if (e.key.keysym.sym == SDLK_F12) { // F12 is screenshot shortcut
				RawImg tempf;
				tempf.data = NULL;
				CreateRaw(&tempf, xres(), yres());
				memcpy(tempf.data, get_image_buffer(), xres()*yres()*4);
				Take_Snapshot(tempf);
				FreeRaw(&tempf);
				}
			if (developer) {
				if (e.key.keysym.sym == SDLK_F11) { // F11 is bilinear filtering toggle shortcut
					bilfilter = !bilfilter;
					}
				if (e.key.keysym.sym == SDLK_F8) {
					switch (fsaa_mode) {
						case FSAA_MODE_NONE:    set_fsaa_mode (FSAA_MODE_4X_LO_Q); break;
						case FSAA_MODE_4X_LO_Q: set_fsaa_mode (FSAA_MODE_4X_HI_Q); break;
						case FSAA_MODE_4X_HI_Q: set_fsaa_mode (FSAA_MODE_NONE); break;
					}
				}
				if (e.key.keysym.sym == SDLK_F5) use_shader = !use_shader;
				if (e.key.keysym.sym == SDLK_F9 ) fov/=1.01;
				if (e.key.keysym.sym == SDLK_F10) fov*=1.01;
// THESE WERE USED TO ADJUST THE AREA CONSTANT
//				if (e.key.keysym.sym == SDLK_F3 ) Area_const *= 0.9;
//				if (e.key.keysym.sym == SDLK_F4 ) Area_const *= 1.1;
			}
			if (e.key.keysym.sym == SDLK_r)	{
				runmode = !runmode;
				if (runmode) speed*=10.0;
					else speed/=10.0;
				}
			}
		}
	keystate = SDL_GetKeyState(NULL);
	if (keystate[SDLK_UP	])	move_camera(	0	);
	if (keystate[SDLK_DOWN	])	move_camera(	M_PI	);
	if (keystate[SDLK_LEFT	])	move_camera(-	M_PI / 2);
	if (keystate[SDLK_RIGHT	])	move_camera(	M_PI / 2);
	if (keystate[SDLK_KP2	])	rotate_camera(	0,	-1*keyboard_sensitivity);
	if (keystate[SDLK_KP4	])	rotate_camera( -1*keyboard_sensitivity,	 0);
	if (keystate[SDLK_KP6	])	rotate_camera( +1*keyboard_sensitivity,	 0);
	if (keystate[SDLK_KP8	])	rotate_camera(	0,	+1*keyboard_sensitivity);
	SDL_GetRelativeMouseState(&deltax, &deltay);
	//SDL_WarpMouse(100, 100);
	rotate_camera(mouse_sensitivity*deltax, -mouse_sensitivity*deltay);
	check_params();
}

void OutroEffect(Uint32 *fb)
{
	int i, j, q=0;

	for (j = 0;j < yres(); j++)
		for (i=0; i< xres(); i++)
			fb[q++] = desaturate(multiplycolor(fb[q], 50000 + rand()%20000), 0.8);
}

void outro(SDL_Surface *p, int exit_code, Uint32 *framebuffer)
{
	int she=0;
#ifdef SHOW_CPU_SPEED
	int cpuspd = GetCPUSpeed();
#endif
#ifdef MAKE_CHECKSUM
	Uint32 cksave;
#endif
	if (SDLFLAGS|SDL_FULLSCREEN) {
		SDL_WM_ToggleFullScreen(p);
		}
#ifdef MAKE_CHECKSUM
	cksave = cksum;
#endif
	load_frame(1);
	render_single_frame(screen, ovr);
#ifdef MAKE_CHECKSUM
	cksum = cksave;
#endif
	vframe--;
	//shader_edge_glow(framebuffer, framebuffer, xres(), yres());
	OutroEffect(framebuffer);
	//shader_gamma_shr(framebuffer, framebuffer, xres(), yres(), 1);
	if (exit_code == 4)
		printxy(p, framebuffer, font0, 45, 32, 0xeef9ff, 0.75, "Test complete.");
		else
		printxy(p, framebuffer, font0, 45, 32, 0xeef9ff, 0.75, "Test complete -- %d loops.", OptionValueInt("--loops"));
	printxy(p, framebuffer, font0, 45, 52, 0xeef9ff, 0.75, "This machine score:");
	printxy(p, framebuffer, font0, 264,52, 0x22ff33, 0.99, "%.2lf FPS.", 1000.0 * (double) vframe / (double) clk);
	printxy(p, framebuffer, font0, 45, 72, 0xeef9ff, 0.75, "Rendering time: %.2lf seconds.", clk / 1000.0);
#ifdef SHOW_CPU_SPEED
	printxy(p, framebuffer, font0, 45, 92, 0xeef9ff, 0.66, "CPU Speed:       MHz");
	printxy(p, framebuffer, font0,166, 92, 0xff3f00, 0.95, "%5d", cpuspd);
#endif
	if (used_threads)
	printxy(p, framebuffer, font0, 45,112, 0xeef9ff, 0.75, "Used two-threaded rendering.");
#ifdef MAKE_CHECKSUM
	printxy(p, framebuffer, font0, 45,100, 0xeef9ff, 0.75, "Expected checksum: %X", get_correct_cksum());
	printxy(p, framebuffer, font0, 45,120, 0xeef9ff, 0.75, "  YOUR   checksum: ");
	if (exit_code == 4)
		if (cksum==get_correct_cksum())
			printxy(p, framebuffer, font0, 254, 120, 0x22ff33, 0.75, "OK");
			else
			printxy(p, framebuffer, font0, 254, 120, 0xff4444, 0.99, "INCORRECT - %X", cksum);
		else
		if (exit_code == 8)
			printxy(p, framebuffer, font0, 254, 120, 0x22ff33, 0.75, "OK (for all loops).");
			else
			printxy(p, framebuffer, font0, 254, 120, 0xff4444, 0.99, "FAILED after loop #%d",
			OptionValueInt("--loops")-loops_remaining);
#endif
	printxy(p, framebuffer, font0, xres()-92, yres()-18, 0x777777, 0.26, "v. 1.03b");
 	Slock(p);
 	memcpy(p->pixels, framebuffer, p->h*p->w*4);
 	Sulock(p);
	SDL_Flip(p);
	while (!she) kbd_tiny_do(&she);
}















int main(int argc, char *argv[])
{ int she=0, loop_mode, loop_iteration=0;
  RawImg introt;
  int ffm = 1, lcr;
  Uint32 SDL_f;
  initcmdline(argc, argv);
  if (OptionExists("--help") || OptionExists("-h")) {
  	DisplayUsage();
	exit(0);
  	}
  if (OptionExists("--benchmark")) {
  	init_yuv_convert(1);
	exit(0);
  	}
 if (OptionExists("--developer")) developer = 1;
 if ( SDL_Init(SDL_INIT_TIMER|SDL_INIT_VIDEO) < 0) {
  	printf("Unable to init SDL: %s\n", SDL_GetError());
	exit(1);
  	}

  atexit(SDL_Quit);
  //-------------------------
  SDL_f = SDLFLAGS;
  if (OptionExists("-w")||OptionExists("--window")||OptionExists("--windowed")) SDL_f &= ~SDL_FULLSCREEN;
  if (OptionExists("-f")||OptionExists("--fullscreen")) SDL_f |= SDL_FULLSCREEN;
  screen = SDL_SetVideoMode(640, 480, 32, SDL_f);
  if (screen == NULL) {
  	printf("Unable to set 640x480 resolution...%s!\n", SDL_GetError());
	exit(1);
  	}

 SDL_WM_SetCaption("Anrieff's Fractal", "");
 if (!LoadBmp(floor_texture, &tex)) {
 	printf("Unable to load main texture (%s), bailing out\n", floor_texture);
	SDL_Delay(500); exit(1) ;
	}
 if (!LoadBmp(default_font, &font0)) { printf("Unable to load default font, bailing out\n"); SDL_Delay(500); exit(1); }
 if (LoadBmp(introtex, &introt)) {
 	PasteRaw(screen, introt, 0, 0);
	PasteRaw(screen, introt, 0, 0); // pasting it twice solves the problem with double-buffering
	}
 precalc();
 FreeRaw(&introt);
 if (OptionExists("--xres")) {
 	SetNewMode(OptionValueInt("--xres"), OptionValueInt("--xres")/4*3);
 	}
 if (xres()!=640) {
 	screen = SDL_SetVideoMode(xres(), yres(), 32, SDL_f);
  	if (screen == NULL) {
  		printf("Unable to set %dx%d resolution...%s!\n", xres(), yres(), SDL_GetError());
		exit(1);
  		}
	mouse_sensitivity *= 640.0 / (double) xres();
	}
 hwaccel_init();
if (!developer && (lcr = LoadContext("data/benchmark.fsv"))>0) {
	//printf("File load success!\n"); fflush(stdout);
	switch (lcr) {
		case SAVLOAD_COORDS:
			file_control = 1;
			kbd_control = 0;
			break;
		case SAVLOAD_OK:
		default:
			file_control = 0;
			kbd_control = 1;
			break;
		}
	} else {
	init_fract_array();
	//SaveContext("benchmark.fsv");
	kbd_control = 1;
	cd_frames = 0x7fffffff;
	}
/* "Game" Loop initialisation */
 set_fsaa_mode(FSAA_STARTING_MODE);
 check_fsaa_param();
 clk = SDL_GetTicks();
#ifdef SHOWFPS
 fpsclk = clk;
#endif
#ifdef BILINEAR_FILTER
 bilfilter = 1;
#endif
 loop_mode = OptionExists("--loop")||OptionExists("--loops");
 if (OptionExists("--loops")) {
 	loops_remaining = OptionValueInt("--loops");
 	}
/* Here follows the Game loop :) */
 lft = SDL_GetTicks();
  while (!she) {
  	camera_moved = 0;
	if (ffm) { ffm = 0; camera_moved = 1; }
	delta = (SDL_GetTicks()-lft)/1000.0;
	light_angle += delta/2;
	/*lx = (int) (200.0 + sin(light_angle)*200.0);
	ly = (int) (100.0 + cos(light_angle)*90.0);
	lz = (int)  (500.0 + cos(light_angle)*250.0);*/
	lft = SDL_GetTicks();
	if (kbd_control) {
		kbd_do(&she);
		} else {
		kbd_tiny_do(&she);
		}
	if (file_control) {
		load_frame(vframe%cd_frames);
		}
#ifdef RECORD
	record_do(bTime());
#endif
//	DrawIt(screen, ovr);
	render_single_frame(screen, ovr);
	if (vframe%cd_frames==0) {
		if (!loop_mode) she = 4;
			else {
				if (!loops_remaining) {
					loops_remaining = -666;
					she = 8;
					}
				loops_remaining--;
#ifdef MAKE_CHECKSUM
				if (cksum!=get_correct_cksum()) {
				    if (loop_iteration>1) {
    					printf("Incorrect checksum at iteration %d: It is %X, should be %X!!\n", loop_iteration,
	   											cksum, get_correct_cksum());
	       				she = 16;
	       				} else set_correct_cksum(cksum);
					}
				cksum = 0;
#endif
				loop_iteration++;
				}
			}
  	}
 clk = SDL_GetTicks() - clk;
 if (!developer && she == 4 || she == 8 || she == 16) {
 	outro(screen, she, get_image_buffer());
 	}
 printf("%d frames drawn for %d milliseconds; this yields %.2f FPS.\n", vframe, clk, 1000.0 * (float) vframe / (float) clk);
 if (developer) prof_statistics();
#ifdef MAKE_CHECKSUM
 printf("Checksum = %X\n", cksum);
#endif
// printf("Area constant: %.6lf\n", Area_const);
 if (ovr) SDL_FreeYUVOverlay(ovr);
 freetextures();
 //free(tex.data);
 FreeRaw(&font0);
 CloseProgram();
 return 0;
}

