/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *-------------------------------------------------------------------------*
 *                                                                         *
 *   rgb2yuv.cpp - handles RGB-to-YUV colorspace convesions.               *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include "MyGlobal.h"


CPUCaps CPU;
int bestmethod=0;

Uint32 checksums[2][2][4] = {
	{ /*AMD*/
/*LinuX*/	{ 0xF781A462, 0xF781A462, 0xE06D540B, 0xFE9A01EB },
/*WindowS*/	{ 0xEBFD23B6, 0xEBFD23B6, 0xD4E8D4D0, 0x03A4FE3F },
	},
	{ /*Intel*/
/*LinuX*/	{ 0xF781A462, 0xF781A462, 0xF781A462, 0xF781A462 },
/*WindowS*/	{ 0xEBFD23B6, 0xEBFD23B6, 0xD4E8D4D0, 0xF33BFC69 }
	}
};

Uint32 csc=0xffffffff;

// these are used for benchmarking purposes
Uint32 fakeRGBBuffer[BENCHSIZE], fakeYUVBuffer[BENCHSIZE];

// the 64 MB bigarray...
Uint32 *bigarray = NULL;

// ConvertRGB2YUV_X86 - convers [count] pixels from [*src] to [*dest]
// transformation is FROM RGB colorspace TO YUY2 packed format
// the only portable function  :) Uses i386 arighmetics only. Fairly precise
// count must be even
void ConvertRGB2YUV_X86(Uint32 *dest, Uint32 *src, size_t count)
{register int r, g, b;
 int r1, g1, b1;
 count /= 2;
 while (count--) {
 	b = (*src) & 0xff;
	g = ((*src)>>8) & 0xff;
	r = ((*src)>>16) & 0xff;
	src++;
 	b1 = (*src) & 0xff;
	g1 = ((*src)>>8) & 0xff;
	r1 = ((*src)>>16) & 0xff;
	(*dest) = /* Y0 */ (((0x00100000 + r * IM11 + g * IM12 + b * IM13)>>16)&0x000000ff)
	        | /* U0 */ (((0x00800000 + r * IM31 + g * IM32 + b * IM33)>> 8)&0x0000ff00)
		| /* V0 */ (((0x00800000 + r * IM21 + g * IM22 + b * IM23)<< 8)&0xff000000)
		| /* Y1 */ (((0x00100000 + r1* IM11 + g1* IM12 + b1* IM13)    )&0x00ff0000)
		;
	dest++;
	src++;
 	}
}


// ConvertRGB2YUV_ASM - convers [count] pixels from [*src] to [*dest]
// transformation is FROM RGB colorspace TO YUY2 packed format
// -- Assembly version of ConvertRGB2YUV_X86. Uses i386 intructions only. No precision loss also.
// count must be even
void ConvertRGB2YUV_X86_ASM(Uint32 *dest, Uint32 *src, size_t count)
{int s[9] = {IM11, IM12, IM13, IM21, IM22, IM23, IM31, IM32, IM33};
 count /= 2;
 while (count--) {
 	__asm __volatile(
	//  Y0 ::
	"mov		%1,		%%ecx\n"
	"movzbl		%%cl,		%%eax\n"
	"imull		8%2\n"
	"xchg		%%eax,		%%ebx\n"
	"movzbl		%%ch,		%%eax\n"
	"ror		$16,		%%ecx\n"
	"imull		4%2\n"
	"add		%%eax,		%%ebx\n"
	"movzbl		%%cl,		%%eax\n"
	"ror		$16,		%%ecx\n"
	"add		$0x100000,	%%ebx\n"
	"imull		%2\n"
	"add		%%ebx,		%%eax\n"
	"shr		$16,		%%eax\n"
	"and		$0xff,		%%eax\n"
	"mov		%%eax,		%0\n"
	//  U0 ::
	"movzbl		%%cl,		%%eax\n"
	"imull		32%2\n"
	"xchg		%%eax,		%%ebx\n"
	"movzbl		%%ch,		%%eax\n"
	"ror		$16,		%%ecx\n"
	"imull		28%2\n"
	"add		%%eax,		%%ebx\n"
	"movzbl		%%cl,		%%eax\n"
	"ror		$16,		%%ecx\n"
	"add		$0x800000,	%%ebx\n"
	"imull		24%2\n"
	"add		%%ebx,		%%eax\n"
	"shr		$8,		%%eax\n"
	"and		$0xff00,	%%eax\n"
	"or 		%%eax,		%0\n"
	//  V0 ::
	"movzbl		%%cl,		%%eax\n"
	"imull		20%2\n"
	"xchg		%%eax,		%%ebx\n"
	"movzbl		%%ch,		%%eax\n"
	"ror		$16,		%%ecx\n"
	"imull		16%2\n"
	"add		%%eax,		%%ebx\n"
	"movzbl		%%cl,		%%eax\n"
	"ror		$16,		%%ecx\n"
	"add		$0x800000,	%%ebx\n"
	"imull		12%2\n"
	"add		%%ebx,		%%eax\n"
	"shl		$8,		%%eax\n"
	"and		$0xff000000,	%%eax\n"
	"or 		%%eax,		%0\n"
	//  Y1 ::
	"mov		4%1,		%%ecx\n"
	"movzbl		%%cl,		%%eax\n"
	"imull		8%2\n"
	"xchg		%%eax,		%%ebx\n"
	"movzbl		%%ch,		%%eax\n"
	"ror		$16,		%%ecx\n"
	"imull		4%2\n"
	"add		%%eax,		%%ebx\n"
	"movzbl		%%cl,		%%eax\n"
	"ror		$16,		%%ecx\n"
	"add		$0x100000,	%%ebx\n"
	"imull		%2\n"
	"add		%%ebx,		%%eax\n"
	"and		$0xff0000,	%%eax\n"
	"or 		%%eax,		%0\n"

	:"=m"(*dest)
	:"m"(*src), "m"(*s)
	:"memory", "eax", "ebx", "ecx", "edx"
	);
	dest++;
	src+=2;
 	}
}

// ConvertRGB2YUV_X86_FPU - convers [count] pixels from [*src] to [*dest] using floating point math
// transformation is FROM RGB colorspace TO YUY2 packed format
// count must be even.
void ConvertRGB2YUV_X86_FPU(Uint32 *dest, Uint32 *src, size_t count)
{
	//float M[9] __attribute__ ((aligned(16))) = {M11, M12, M13, M21, M22, M23, M31, M32, M33};
	float M[9] __attribute__ ((aligned(16))) = {M13, M33, M23, M11, M31, M21, M12, M32, M22};
	int   a[2] __attribute__ ((aligned(16))) = {16, 128};
	Uint16 cwsave, mycw;

	count /= 2;
	// lower FPU precision to single:
	__asm __volatile("	fstcw	%0\n":"=m"(cwsave)::"memory"); // get conrol word
	mycw = cwsave & (~0x0300); // set single precision
	__asm __volatile("	fldcw	%0\n"::"m"(mycw):"memory"); // put control word
	while (count--) {
		__asm __volatile (

//get B0, R0, G0:
		"movl	%1,		%%eax\n"
		"andl	$0xff00ff,	%1\n"
		"mov	$0x00ff00,	%%edx\n"
		"filds	%1\n"
		"and	%%eax,	%%edx\n"
		"shrl	$16,	%1\n"
		"shrl	$8,	%%edx\n"
		"filds	%1\n"
		"mov	%%edx,	%1\n"
		"filds	%1\n"
// st(0) = g0; st(1) = r0; st(2) = b0

// results will be: st -> V0, st(1) -> U0, st(2) -> Y0

// B0:
		"fld	%%st(2)\n"
		"fmul	%2\n"
		"fld	%%st(3)\n"
		"fmul	4%2\n"
		"fld	%%st(4)\n"
		"ffree	%%st(5)\n"
		"fmul	8%2\n"

// R0:
		"fld	%%st(4)\n"
		"fmul	12%2\n"
		"fld	%%st(5)\n"
		"fmul	16%2\n"
		"fld	%%st(6)\n"
		"fmul	20%2\n"
		"ffree	%%st(7)\n"

		"faddp	%%st,	%%st(3)\n"
		"faddp	%%st,	%%st(3)\n"
		"faddp	%%st,	%%st(3)\n"
// G0:
		"fld	%%st(3)\n"
		"fmul	24%2\n"
		"fld	%%st(4)\n"
		"fmul	28%2\n"
		"fld	%%st(5)\n"
		"fmul	32%2\n"
		"ffree	%%st(6)\n"

		"faddp	%%st,	%%st(3)\n"
		"faddp	%%st,	%%st(3)\n"
		"faddp	%%st,	%%st(3)\n"

		"fiaddl 4%3\n"

// ST(0) = V0; ST(1) = U0; ST(2) = Y0
		"fistps	%0\n"
		"fiaddl	4%3\n"
		"fistps	%1\n"
		"fiaddl	%3\n"
		"shll	$24,	%0\n"

		"fistps	%0\n"
		"movzx	%1,	%%edx\n"
		"shl	$8,	%%edx\n"
		"orl	%%edx,	%0\n"

// %0 = V0 00 U0 Y0
		"movl	%%eax,	%1\n"

		"fildl	%3\n"
		"flds	24%2\n"
		"flds	12%2\n"
		"flds	%2\n"

		"movl	4%1,	%%eax\n"
		"movl	$0x00ff00,	%%edx\n"
		"andl	$0xff00ff,	4%1\n"
		"andl	%%eax,	%%edx\n"
		"fimuls 4%1\n"
		"shrl	$16,	4%1\n"
		"faddp	%%st,	%%st(3)\n"
		"shrl	$8,	%%edx\n"
		"fimuls	4%1\n"
		"faddp	%%st,	%%st(2)\n"
		"movl	%%edx,	4%1\n"
		"fimuls 4%1\n"

		"faddp\n"

		"fistpl	4%1\n"
		"movl	4%1,	%%edx\n"
		"shll	$16,	%%edx\n"
		"orl	%%edx,	%0\n"
		"movl	%%eax,	4%1\n"

		:"=m"(*dest), "=m"(*src)
		:"m"(*M), "m"(*a)
		:"memory", "eax", "edx"
		);
		dest++; src+=2;
	}
	__asm __volatile ("	fldcw	%0\n"::"m"(cwsave):"memory");
}

// ConvertRGB2YUV_MMX - convers [count] pixels from [*src] to [*dest] using MMX
// transformation is FROM RGB colorspace TO YUY2 packed format
// uses tightly optimized (pairable) MMX code
// should work significantly faster on PIII & P4. (And of course AthlonXP). It's not such a big deal on a PPro and PII
// there is a small precision cost, which should be unnoticable
// count must be even.

void ConvertRGB2YUV_MMX(Uint32 *dest, Uint32 *src, size_t count)
{Sint16 pmm3[4]={SIM13, SIM33, SIM13, SIM23};
 Sint16 pmm4[4]={SIM12, SIM32, SIM12, SIM22};
 Sint16 pmm5[4]={SIM11, SIM31, SIM11, SIM21};
 Uint16 pmm6[4]={4096, 32768, 4096, 32768};
 Sint16 pmm7[4]={0, 0, 0, 0};


 count /= 2;
 // load some useful values
 __asm __volatile(
 "movq %0, %%mm3\n"
 "movq %1, %%mm4\n"
 "movq %2, %%mm5\n"
 "movq %3, %%mm7\n"
 "movq %4, %%mm6\n"
 ::"m"(*pmm3), "m"(*pmm4), "m"(*pmm5), "m"(*pmm7), "m"(*pmm6):"memory"
 );
 while (count--) {
	__asm __volatile(
	// get the current pixel in ebx, the next in edx
	"mov	%1, 	%%ebx\n"
	"mov	4%1,	%%edx\n"
	// prepare mm0 - fill with (b0, b0, b1, b0)
	"movb	%%bl,	%%ah\n"
	"movb	%%dl, 	%%al\n"
	"shl	$16,	%%eax\n"
	"movb	%%bl,	%%al\n"
	"movb	%%bl,	%%ah\n"
	"movd	%%eax,	%%mm0\n"

	"shr	$8,	%%ebx\n"
	"shr	$8,	%%edx\n"

	// prepare mm1 - fill with (g0, g0, g1, g0)
	"movb	%%bl,	%%ah\n"
	"movb	%%dl, 	%%al\n"
	"shl	$16,	%%eax\n"
	"movb	%%bl,	%%al\n"
	"movb	%%bl,	%%ah\n"
	"movd	%%eax,	%%mm1\n"

	"shr	$8,	%%ebx\n"
	"shr	$8,	%%edx\n"

	// prepare mm2 - fill with (r0, r0, r1, r0)
	"movb	%%bl,	%%ah\n"
	"movb	%%dl, 	%%al\n"
	"shl	$16,	%%eax\n"
	"movb	%%bl,	%%al\n"
	"movb	%%bl,	%%ah\n"
	"movd	%%eax,	%%mm2\n"

	// do the actual math:

	"punpcklbw 	%%mm7,	%%mm0\n"
	"punpcklbw 	%%mm7,	%%mm1\n"
	"punpcklbw 	%%mm7,	%%mm2\n"

	"pmullw 	%%mm3,	%%mm0\n"
	"pmullw 	%%mm4,	%%mm1\n"
	"pmullw 	%%mm5,	%%mm2\n"

	"paddw		%%mm1,	%%mm0\n"
	"paddw		%%mm6,	%%mm2\n"
	"paddw		%%mm2,	%%mm0\n"
	"psrlw		$8,	%%mm0\n"

	"packuswb	%%mm7,	%%mm0\n"

	"movd		%%mm0,	%0\n"

	:"=m"(*dest)
	:"m"(*src)
	:"memory", "eax", "ebx", "edx"
	);
 	dest++;
	src+=2;
 	}
 __asm __volatile("emms");
}

// ConvertRGB2YUV_MMX2 - convers [count] pixels from [*src] to [*dest] using MMX2
// transformation is FROM RGB colorspace TO YUY2 packed format
// uses tightly optimized (pairable) MMX code
// count must be even.

void ConvertRGB2YUV_MMX2(Uint32 *dest, Uint32 *src, size_t count)
{Sint16 pmm3[4]={SIM13, SIM33, SIM13, SIM23};
 Sint16 pmm4[4]={SIM12, SIM32, SIM12, SIM22};
 Sint16 pmm5[4]={SIM11, SIM31, SIM11, SIM21};
 Uint16 pmm6[4]={4096, 32768, 4096, 32768};
 Uint16 pmm7[4]={0xffff, 0x00ff, 0xffff, 0x00ff};
 //Sint16 pmm7[4]={0, 0, 0, 0};

 count /= 2;

 // load some constants:
 __asm __volatile (
 "	pxor	%%mm7,	%%mm7\n"
 "	movq	%0,	%%mm6\n"
 ::"m"(*pmm7):"memory"
 );

 while (count--) {
 	__asm __volatile (
	"	movq	%1,	%%mm0\n"
	"	pand	%%mm6,	%%mm0\n"
	"	pshufw	$0x0e,	%%mm0,	%%mm1\n"
// mm0 contains pixel 1 and pixel 2
// low dword of mm1 contains pixel 2

	"	punpcklbw	%%mm7,	%%mm0\n"
	"	punpcklbw	%%mm7,	%%mm1\n"

// mm0 = (b0 g0 r0 00) (words)
// mm1 = (b1 g1 r1 00) (words)

// SOME REAALY WEIRD SHUFFLES::: :)
	"	pshufw	$0x30,	%%mm0,	%%mm2\n" // 00110000
	"	pshufw	$0xcf,	%%mm1,	%%mm3\n" // 11001111
	"	pshufw	$0x75,	%%mm0,	%%mm4\n" // 01110101
	"	pshufw	$0xdf,	%%mm1,	%%mm5\n" // 11011111
	"	pshufw	$0xba,	%%mm0,	%%mm0\n" // 10111010
	"	pshufw	$0xef,	%%mm1,	%%mm1\n" // 11101111
// ends up like this:
	"	por	%%mm3,	%%mm2\n" // mm2 - b0 b0 b1 b0
	"	por	%%mm5,	%%mm4\n" // mm4 - g0 g0 g1 g0
	"	por	%%mm1,	%%mm0\n" // mm0 - r0 r0 r1 r0

	"	pmullw	%2,	%%mm2\n" // mm2 *= M13 M33 M13 M23
	"	pmullw	%3,	%%mm4\n" // mm4 *= M12 M32 M12 M22
	"	pmullw	%4,	%%mm0\n" // mm0 *= M11 M31 M11 M21

	"	paddw	%5,	%%mm2\n" // mm2 += 16  128  16 128
	"	paddw	%%mm4,	%%mm0\n"
	"	paddw	%%mm2,	%%mm0\n"
	"	psrlw	$8,	%%mm0\n" // mm0 = Y0 U0 Y1 V0

	"	packuswb	%%mm7,	%%mm0\n"
	"	movd	%%mm0,	%0\n"


	:"=m"(*dest)
	:"m"(*src), "m"(*pmm3), "m"(*pmm4), "m"(*pmm5), "m"(*pmm6)
	:"memory"
	);
 	src+=2;
	dest++;
 }

 __asm __volatile("emms");
}


// ConvertRGB2YUV_SSE - convers [count] pixels from [*src] to [*dest] using MMX2 and SSE
// transformation is FROM RGB colorspace TO YUY2 packed format
// count must be divisable by four! (twice unrolled loop)

void ConvertRGB2YUV_SSE(Uint32 *dest, Uint32 *src, size_t count)
{
	float __attribute__ ((aligned (16))) cxmm4[4] = { M13, M33, M13, M23 };
	float __attribute__ ((aligned (16))) cxmm5[4] = { M12, M32, M12, M22 };
	float __attribute__ ((aligned (16))) cxmm6[4] = { M11, M31, M11, M21 };
	float __attribute__ ((aligned (16))) cxmm7[4] = {  16, 128,  16, 128 };
	Uint16 cmm6[4] = {0xffff, 0x00ff, 0xffff, 0x00ff};
	count /= 4;
	__asm __volatile ("pxor	%%mm7,	%%mm7\n":::"memory");
	while (count -- ) {

/*
	Assembly Implementation:
	get 16 bytes per cycle (4 points) to produce 8 bytes of YUV results
	i.e. the conversion loop is unrolled once. Proved to be 15% faster than
	the non-unrolled variant on an Athlon XP
*/
		__asm __volatile (
		"	movq	8%1,	%%mm0\n"
		"	movq	%1,	%%mm4\n"
		"	pand	%6,	%%mm0\n"
		"	pand	%6,	%%mm4\n"
		"	pshufw	$0x0e,	%%mm0,	%%mm1\n"
		"	pshufw	$0x0e,	%%mm4,	%%mm5\n"
// hint hardware to prefetch next 4 points:
		"	prefetchnta	16%1\n"

		"	punpcklbw	%%mm7,	%%mm0\n"
		"	punpcklbw	%%mm7,	%%mm4\n"
		"	punpcklbw	%%mm7,	%%mm1\n"
		"	punpcklbw	%%mm7,	%%mm5\n"

// mm0 = (b0, g0, r0, 00) (words)
// mm1 = (b1, g1, r1, 00) (words)
///---------------------------------------
// mm4 = (b2, g2, r2, 00) (words)
// mm5 = (b3, g3, r3, 00) (words)

		"	pshufw	$0xcc,	%%mm0,	%%mm2\n"
		"	pshufw	$0xcc,	%%mm4,	%%mm6\n"
		"	pshufw	$0xcc,	%%mm1,	%%mm3\n"
		"	pshufw	$0xcc,	%%mm5,	%%mm7\n"

		"	cvtpi2ps	%%mm2,	%%xmm1\n"
		"	cvtpi2ps	%%mm6,	%%xmm5\n"
		"	cvtpi2ps	%%mm3,	%%xmm0\n"
		"	cvtpi2ps	%%mm7,	%%xmm4\n"

		"	pshufw	$0xdd,	%%mm0,	%%mm2\n"
		"	pshufw	$0xdd,	%%mm1,	%%mm3\n"
		"	pshufw	$0xdd,	%%mm4,	%%mm6\n"
		"	pshufw	$0xdd,	%%mm5,	%%mm7\n"

		"	movlhps	%%xmm0,	%%xmm1\n"
		"	movlhps	%%xmm4,	%%xmm5\n"

		"	pshufw	$0xee,	%%mm0,	%%mm0\n"
		"	pshufw	$0xee,	%%mm1,	%%mm1\n"
		"	pshufw	$0xee,	%%mm4,	%%mm4\n"
		"	pshufw	$0xee,	%%mm5,	%%mm5\n"

		"	mulps	%2,	%%xmm1\n"
		"	mulps	%2,	%%xmm5\n"

		"	cvtpi2ps	%%mm2,	%%xmm2\n"
		"	cvtpi2ps	%%mm3,	%%xmm3\n"
		"	cvtpi2ps	%%mm6,	%%xmm6\n"
		"	cvtpi2ps	%%mm7,	%%xmm7\n"

		"	movlhps	%%xmm3,	%%xmm2\n"
		"	movlhps	%%xmm7,	%%xmm6\n"

		"	cvtpi2ps	%%mm0,	%%xmm3\n"
		"	cvtpi2ps	%%mm1,	%%xmm0\n"
		"	cvtpi2ps	%%mm4,	%%xmm7\n"
		"	cvtpi2ps	%%mm5,	%%xmm4\n"

		"	mulps	%3,	%%xmm2\n"
		"	mulps	%3,	%%xmm6\n"

		"	movlhps	%%xmm0,	%%xmm3\n"
		"	movlhps	%%xmm4,	%%xmm7\n"

		"	movaps	%5,	%%xmm0\n"
		"	movaps	%%xmm0,	%%xmm4\n"

		"	mulps	%4,	%%xmm3\n"
		"	mulps	%4,	%%xmm7\n"

		"	pxor	%%mm7,	%%mm7\n"

		"	addps	%%xmm1,	%%xmm0\n"
		"	addps	%%xmm5,	%%xmm4\n"
		"	addps	%%xmm3,	%%xmm2\n"
		"	addps	%%xmm7,	%%xmm6\n"
		"	addps	%%xmm2,	%%xmm0\n"
		"	addps	%%xmm6,	%%xmm4\n"

		"	movhlps	%%xmm0,	%%xmm1\n"
		"	movhlps	%%xmm4,	%%xmm5\n"

		"	cvtps2pi	%%xmm0,	%%mm0\n"
		"	cvtps2pi	%%xmm4,	%%mm4\n"
		"	cvtps2pi	%%xmm1,	%%mm1\n"
		"	cvtps2pi	%%xmm5,	%%mm5\n"

		"	packssdw	%%mm1,	%%mm0\n"
		"	packssdw	%%mm5,	%%mm4\n"
		"	packuswb	%%mm7,	%%mm0\n"
		"	packuswb	%%mm7,	%%mm4\n"

		//"	movd	%%mm0,	%0\n"
		"	psllq	$32,	%%mm0\n"
		"	por	%%mm0,	%%mm4\n"
		//"	movd	%%mm4,	4%0\n"
// we prefer movntq while it minimizes cache pollution:
		"	movntq	%%mm4,	%0\n"

		:"=m"(*dest)
		:"m"(*src), "m"(*cxmm4), "m"(*cxmm5), "m"(*cxmm6), "m"(*cxmm7), "m"(*cmm6)
		:"memory"
		);
		dest+=2;
		src+=4;
	}
	__asm __volatile ("emms");
}


// ConvertRGB2YUV_MEM - convers [count] pixels from [*src] to [*dest] using bigarray
// transformation is FROM RGB colorspace TO YUY2 packed format
// This should be portable. Consumes hell a lot of memory (64MB). Using it on a unitialized array could be catastrophic
// Really NO precision loss - always correct results.
// count must be even
void ConvertRGB2YUV_MEM(Uint32 *dest, Uint32 *src, size_t count)
{
 count /= 2;
 while (count--) {
 	(*dest++) = bigarray[(*src)&0xffffff] + ((bigarray[(*(src+1))&0xffffff]<<16)&0xff0000);
	src+=2;
 	}
}


// Master function, switches to the fastest available conversion routine
/***
 *** you may wonder why I've included the X86 version at all (it'll be always slower than the X86_ASM version).
 *** well, someday I may decide to make this whole thing really portable :))
 *************************************************************************************************************/

void ConvertRGB2YUV(Uint32 *dest, Uint32 *src, size_t count)
{
 switch (bestmethod) {
 	case USE_X86: ConvertRGB2YUV_X86(dest, src, count); break;
	case USE_MMX: ConvertRGB2YUV_MMX(dest, src, count); break;
	case USE_X86_ASM: ConvertRGB2YUV_X86_ASM(dest, src, count); break;
	case USE_X86_FPU: ConvertRGB2YUV_X86_FPU(dest, src, count); break;
	case USE_MMX2: ConvertRGB2YUV_MMX2(dest, src, count); break;
	case USE_SSE: ConvertRGB2YUV_SSE(dest, src, count); break;
	case USE_BIGMEM: ConvertRGB2YUV_MEM(dest, src, count); break;
 	}
}


// full YUV conversion ... when we precalculate we could be as slow (but precise) as hell :-)
static inline Uint32 fRGB2YUV(int r, int g, int b)
{int Y, U, V;
 Y = /*clamp(*/(int) (16+ (M11 * r + M12 * g + M13 * b))/*)*/;
 V = /*clamp(*/(int) (128+(M21 * r + M22 * g + M23 * b))/*)*/;
 U = /*clamp(*/(int) (128+(M31 * r + M32 * g + M33 * b))/*)*/;
 return (Y | (U<<8) | (V<<24));
}

// some string funcions:

// returns the position of the first occurence of PATTERN in S. If PATTERN cannot be found within S, pos returns -1
int pos(const char *pattern, const char *s)
{unsigned int j, i, ll, f;

 ll = strlen(pattern);
 for (j=0;j<strlen(s);j++) {
	f = 1;
	for (i=0;i<ll;i++)
		if (pattern[i]!=s[j+i]) {
			f = 0; break;
			}
	if (f) return j;
 	}
 return -1;
}

/***************************************************************
 * void GetCPUCaps - returns some flags about the processor's  *
 *                   capabilities                              *
 *-------------------------------------------------------------*
 *                                                             *
 * The Linux version relies heavily on the /proc filesystem :) *
 ***************************************************************/
#ifdef linux
void GetCPUCaps(CPUCaps *a)
{FILE *f;
 int i, k, memtotal;
 char buf[256];

// nullificate the CPU capabilities:
 a->hasmmx = 0;
 a->hasmmx2= 0;
 a->hassse = 0;
 a->largemem=0;
 if (OptionExists("--has-mmx")) a->hasmmx=1;
 if (OptionExists("--has-sse")) a->hassse=1;
 if (OptionExists("--has-mmx2")) a->hasmmx2=1;
 if ((f=fopen("/proc/cpuinfo", "rt"))==NULL) {
 	printf("Can't open /proc/cpuinfo... falling back on defaut (hardcoded) config (MMX=0, BIGMEM=0)\n");
	return;
 	}
while (!feof(f)) {
	fgets(buf, 256, f);
	if (pos("flags", buf)!=-1) {
		// scan for cpu features
		if (pos("mmx", buf)!=-1) // do we have MMX
			a->hasmmx = 1;
		if (pos("sse", buf)!=-1) // do we have SIMD instructions?
			a->hassse = 1;
		if (pos("mmxext", buf)!=-1) // do we have the extended MMX instruction set?
			a->hasmmx2 =1;
		}
	}
 fclose(f);
 if (OptionExists("--no-mmx")) a->hasmmx=0;
 if (OptionExists("--no-sse")) a->hassse=0;
 if (OptionExists("--no-mmx2")) a->hasmmx2=0;
 if ((f=fopen("/proc/meminfo", "rt"))==NULL) {
 	printf("Can't get memory info - assuming no large memory allocation.");
	a->largemem = 0;
	return;
 	}
 while (!feof(f)) {
 	fgets(buf, 256, f);
	if (pos("MemTotal", buf)) {
		i = 0;
		while (buf[i]<'0'||buf[i]>'9') i++;
		k = i;
		while (buf[k]>='0'&&buf[k]<='9') k++;
		buf[k] = 0;
		sscanf(buf+i, "%d", &memtotal);
		// fixme : assuming memory is given in kilobytes:
		memtotal /= 1024;
		a->largemem = (memtotal >= MEMORY_REQUIREMENT);
		}
 	}
 fclose(f);
 if (OptionExists("--no-mem")) a->largemem=0;
}
#else
/***************************************************************
 * void GetCPUCaps - returns some flags about the processor's  *
 *                   capabilities                              *
 *-------------------------------------------------------------*
 * This is the Windows version, requires CPUID. Should work on *
 * both AMD and Intel CPUs. Code taken from the internet:      *
 *  (looks very ugly :)                                        *
 * http://www.tommesani.com/DetectingMMXSSE.html               *
 ***************************************************************/
void GetCPUCaps(CPUCaps *a)
{int CPUZ=0, CPUEXT=0, ECS=0;

 __asm __volatile (
 "          mov     $1,    %%eax \n"
 "          cpuid \n"
 "          mov     %%edx,  %0\n"
 "	    mov     $0x80000000, %%eax\n"
 "          cpuid\n"
 "          cmp     $0x80000001, %%eax\n"
 "          jb      not_supported\n"

 "          mov     $0x80000001, %%eax\n"
 "          cpuid\n"
 "          mov     %%edx,       %1\n"
 "          movl     $1,          %2\n"
 "          jmp     ende\n"

 "          not_supported:\n"
 "          movl     $0,          %2\n"
 "          ende:\n"
 :"=m"(CPUZ), "=m"(CPUEXT), "=m"(ECS)
 :
 :"eax", "ebx", "ecx", "edx"
 );

 a->hasmmx = (CPUZ&MMX_MASK?1:0);
 a->hassse = (CPUZ&SSE_MASK?1:0);
 if (a->hassse) a->hasmmx2 = 1;
     else {
     if (ECS && (CPUEXT&MMX2_MASK)) a->hasmmx2=1;
                               else a->hasmmx2=0;
     }
 if (OptionExists("--no-sse")) 	 a->hassse	= 0;
 if (OptionExists("--no-mmx")) 	 a->hasmmx	= 0;
 if (OptionExists("--no-mmx2"))  a->hasmmx2	= 0;
 if (OptionExists("--has-sse"))  a->hassse	= 1;
 if (OptionExists("--has-mmx"))  a->hasmmx	= 1;
 if (OptionExists("--has-mmx2")) a->hasmmx2	= 1;
 a->largemem = 0;
 //printf("Detected CPU features: mmx=%d, mmx2=%d, sse=%d, bigmem=%d\n", a->hasmmx, a->hasmmx2, a->hassse, a->largemem);
}
#endif

int is_intel(void) // for benchmarking only:
{
	int result;

	__asm __volatile(
		"xorl	%%eax, 	%%eax\n"
		"cpuid\n"
		"movl	%%ebx,  %0\n"
	:"=m"(result)
	:
	:"eax", "ebx", "ecx", "edx"
	);
	return (result == INTEL_EBX);
}


// Benchmarks a RGB-to-YUY2 conversion function.
// Exact results may depend on the test data (fakeRGBBuffer) and the BENCHSIZE
// random data is a killer for ConvertRGB2YUV_MEM, as it performs about 6-10 times worse, but works fine on real data.
void BenchmarkFunction( __convert_fn_t fu, int code, char *method, int *maxfps, Uint32 timetorun)
{int cfps, frames=0;
 Uint32 ticks;
 printf("  ConvertRGB2YUV%s: ", method);
 fu(fakeYUVBuffer, fakeRGBBuffer, BENCHSIZE);
 ticks = SDL_GetTicks();
 while (ticks == SDL_GetTicks())
 	/*nothing*/;
 ticks = SDL_GetTicks();
 while (SDL_GetTicks()-ticks<timetorun) {
 	fu(fakeYUVBuffer, fakeRGBBuffer, BENCHSIZE);
 	frames++;
 	}
 ticks = SDL_GetTicks() - ticks;
 cfps = frames * 1000 / ticks;
 printf("%d fps\n", cfps);
 if (cfps > *maxfps) { *maxfps = cfps; bestmethod = code; }
 fflush(stdout);
}

void init_yuv_convert(int benchmark)
{int i;
 RawImg bm;
 bm.data = NULL;
 // get CPU capabilities
 GetCPUCaps(&CPU);
 if (OptionExists("--force-mem"))
 	CPU.largemem = 1;
 // disable all benchmarking and algorithm selection if we have some input on the command line
 if (	OptionExists("--use-x86")    ||OptionExists("--use-x86asm")||
	OptionExists("--use-mmx")    ||OptionExists("--use-mem")   ||
	OptionExists("--use-x86-fpu")||OptionExists("--use-sse")   ||
	OptionExists("--use-mmx2")) {
 	bestmethod = -1;
	if (OptionExists("--use-x86")) bestmethod = USE_X86;
	if (OptionExists("--use-x86asm")) bestmethod = USE_X86_ASM;
	if (OptionExists("--use-mmx") && CPU.hasmmx) bestmethod = USE_MMX;
	if (OptionExists("--use-x86-fpu")) bestmethod = USE_X86_FPU;
	if (OptionExists("--use-sse") && CPU.hasmmx2 && CPU.hassse) bestmethod = USE_SSE;
	if (OptionExists("--use-mmx2") && CPU.hasmmx2) bestmethod = USE_MMX2;
	if (bestmethod!=-1) return;
 	}

 // put random data in the RGB buffer... this will produce more realistic results when benchmarking
 // all convesion functions that utilize lookup tables (will eliminate cache on large tables)
 for (i=0;i<BENCHSIZE;i++)
 	fakeRGBBuffer[i] = rand()%0x1000000;
 // it would be better if we can load a pre-rendered image to test the bigarray method on some 'real' data...
 if (LoadBmp("data/fract_benchmark.bmp", &bm)) {
	memcpy(fakeRGBBuffer, bm.data, bm.x*bm.y*4);
 	FreeRaw(&bm);
 	}
 // if we have enough memory, then create the big lookup array.
 if (CPU.largemem) {
 	if ((bigarray=(Uint32 *) malloc(0x4000000))==NULL) {
		printf("Cannot allocate memory for fast lookup table! Won't use this method!\n");
		CPU.largemem = 0;
		} else
		{
		printf("precalculating bigarray...");
		fflush(stdout);
		for (i=0;i<0x1000000;i++)
			bigarray[i] = fRGB2YUV((i & 0xff0000)>>16, (i & 0xff00) >> 8, (i & 0xff));
		printf("done\n");
		}

 	}
 yuv_benchmark(benchmark);
}

Uint32 get_correct_cksum(void)
{
	int os=1, iss;
	CPUCaps a;
	if (csc!=0xffffffff) return csc;
#ifdef linux
	os=0;
#endif
	GetCPUCaps(&a);
	iss = 0;
	if (a.hasmmx) iss++;
	if (a.hasmmx2) iss++;
	if (a.hassse) iss++;
	csc = checksums[is_intel()][os][iss];
	return csc;
}

void set_correct_cksum(Uint32 c)
{
    csc = c;
}

void rgb2yuv_close(void)
{
 if (bigarray)
 	free(bigarray);
}

// the parameter is to specify --benchmark mode. If set, all conversion functions are run one minute unconditionally.
void yuv_benchmark(int benchmark)
{int maxfps = 0;
 int timetorun;
 timetorun = (benchmark?BENCH_LARGE_TICKS:BENCHTICKS);
 printf("Benchmarking RGB-to-YUV conversion functions:\n");
 	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_X86, USE_X86, "_X86", &maxfps, timetorun);
	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_X86_ASM, USE_X86_ASM, "_X86_ASM", &maxfps, timetorun);
	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_X86_FPU, USE_X86_FPU, "_X86_FPU", &maxfps, timetorun);
 if (CPU.hasmmx)
 	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_MMX, USE_MMX, "_MMX", &maxfps, timetorun);
 if (CPU.hasmmx2)
 	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_MMX2, USE_MMX2, "_MMX2", &maxfps, timetorun);
 if (CPU.hasmmx2 && CPU.hassse) {
 	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_SSE, USE_SSE, "_SSE", &maxfps, timetorun);
	}
 if (CPU.largemem)
 	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_MEM, USE_BIGMEM, "_MEM", &maxfps, timetorun);
 if (OptionExists("--use-mem") && CPU.largemem) bestmethod = USE_BIGMEM;
}
