/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifndef __SHADERS_H__
#define __SHADERS_H__

#include <SDL/SDL.h>

// if defined, this will reduce the size of the shaders.cpp framebuffers to save some memory
#define SAVE_MEM

#ifdef SAVE_MEM
#define SHADER_MAXX 8
#define SHADER_MAXY  8
#define MAX_FFT_SIZE 8
#else // full config:
#define SHADER_MAXX 1280
#define SHADER_MAXY  960
#define MAX_FFT_SIZE 1024
#endif // SAVE_MEM

#define MA3X_MAX_SIZE 8


typedef struct{
float re, im;
}complex;

typedef struct {
	int coeff[MA3X_MAX_SIZE][MA3X_MAX_SIZE];
	int n;
}ConvolveMatrix;

extern ConvolveMatrix pshop_ma3x_3, pshop_ma3x_5, pshop_ma3x_7;
extern ConvolveMatrix blur_ma3x_3, blur_ma3x_5;

void effects_init(void);
void shaders_close(void);

void fft_2d_complex(complex *src, complex *dst, int n, int isi);

/* Does FFT on the whole image, applies the given filter and does an inverted FFT. Quite slow. */
// CAN DO THE EFFECT IN-PLACE
int shader_FFT_Filter(Uint32 *src, Uint32 *dst, int resx, int resy);

/* Does convolution with the given ConvolveMatrix. Matrices may be of size 3x3, 5x5 and 7x7 with the
   3x3 and 5x5 version being exceptably fast on MMX enabled machines*/
// CAN'T DO THE EFFECT IN-PLACE
void shader_convolution(Uint32 *src, Uint32 *dest, int resx, int resy, ConvolveMatrix *M);




/* Does an Edge Glow effect using convolution with 5x5 matrix and 3x3 matrix blur */
// CAN DO THE EFFECT IN-PLACE
void shader_edge_glow(Uint32 *src, Uint32 *dest, int resx, int resy);

// does a simple matrix blur. blur_size indicates the amount of blur can be 3, 5 or 7
void shader_blur(Uint32 *src, Uint32 *dest, int resx, int resy, int blur_size);

void shader_gamma_shl(Uint32 *src, Uint32 *dest, int resx, int resy, int shift);
void shader_gamma_shr(Uint32 *src, Uint32 *dest, int resx, int resy, int shift);


#endif//__SHADERS_H__
