/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *-------------------------------------------------------------------------*
 *                                                                         *
 *   sphere.cpp - handles sphere creation (the fractal formula) and        *
 *                sorting                                                  *
 *                                                                         *
 ***************************************************************************/

#include "MyGlobal.h"
#include <math.h>
#include <stdlib.h>

sphere *a, t[MAX_SPHERES];
int n;
int defaultflags=0;

AniSphere AS[100];
int anicnt;
const char *earthmap = "data/1440.bmp";

// decompose a RGB value to its components
void setcolor(sphere *a, int col)
{
 a->b = (col & 0xff);
 col >>= 8;
 a->g = (col & 0xff);
 col >>= 8;
 a->r = (col & 0xff);
}

int rand_color(void)
{
    return (0x606060 + ((rand()%0x1000000)&0x7f7f7f));
}
int isec = 0;
int intersect(sphere *a)
{
    int i;

    for (i=0;i<n;i++)
        if (pith0(a[i].x, a[i].y, a[i].z, a[n].x, a[n].y, a[n].z) < a[i].d + a[n].d) {
            printf("%d isect %d\n", i, n);
            fflush(stdout);
            isec++;
            if (isec > 100)  {
                printf("A:\n");
                print_sphere_info(stdout, a+i);
                printf("B:\n");
                print_sphere_info(stdout, a+n);
                exit(1);
                }
            return 1;
            }

    return 0;
}

void set_params(sphere *a, double x, double y, double z, double d, double xi, double yi, double zi)
{
 a->x = x; a->y = y; a->z = z; a->d = d; a->xi = xi; a->yi = yi; a->zi = zi;
 a->mass = 1.0;
 a->refl = 0.25;
 setcolor(a, rand()%0x1000000);
 a->flags = defaultflags;
 a->AniIndex = 0;
}

// the recursive routine to create the fractal
void recu(sphere *a, double x, double y, double z, double r, int level)
{
 if (level == 0) {
 	a[n].x = x;
	a[n].y = y;
	a[n].z = z;
	a[n].d = r;
	a[n].flags = RAYTRACED|RECURSIVE|STATIC|COLLIDEABLE;
	a[n].mass = 9999;
	//a->tex = (RawImg *) malloc(sizeof(RawImg));
	//LoadBmp(earthmap, a->tex);
	a[n].refl = 0.2;
	if (r<10)
	setcolor(a+n,multiplycolorf(KOLOR1, r/8.5) + multiplycolorf(KOLOR2, (1-r/8.5)));
	else setcolor(a+n, KOLOR3);
	n++;
	return;
 	}
 recu(a, x       , y       , z+0.4*r, r*0.5, level-1);
 recu(a, x       , y       , z-0.7*r, r*0.3, level-1);
 recu(a, x+0.7*r , y       , z      , r*0.3, level-1);
 recu(a, x-0.7*r , y       , z      , r*0.3, level-1);
 recu(a, x       , y+0.7*r , z      , r*0.3, level-1);
 recu(a, x       , y-0.7*r , z      , r*0.3, level-1);
}

void create_fract_array(sphere *a, int *count)
{int j;
 double M_SPEED=2.5;
 double oo[3];

 if (OptionExists("--speed"))
    M_SPEED = OptionValueFloat("--speed");
 n = 0;
 srand(5);
 recu(a, 200, 95, 330, 26, 0);
 for (j=0;j<10;j++) {
    do {
     	a[n].d = rand()%4 + 6;
    	a[n].refl = 0.28;
    	a[n].flags = RAYTRACED|RECURSIVE|ANIMATED|COLLIDEABLE;
    	a[n].AniIndex = n;
    	a[n].mass = 1;
    	setcolor(a+n, rand_color());
    	AS[n].alpha = (double) (rand()%628) * 0.01;
    	//AS[n].beta =  (double) (rand()%314) * 0.01 - M_PI/2.0;
    	AS[n].beta = (double) (rand()%50)*0.01 - 0.25;
    	AS[n].speed = (0.2 + (rand()%20)*0.1)*0.11*M_SPEED;
    	AS[n].radius = 36 + rand()%60;
    	AS[n].cx = a[0].x;
    	AS[n].cy = a[0].y;
    	AS[n].cz = a[0].z;
    	calculate_XYZ(a+n, 0, oo);
//	calculate_XYZ(a+n, (rand()%1000) * 0.01, oo);
    	a[n].x = oo[0]; a[n].y = oo[1]; a[n].z = oo[2];
    	} while (intersect(a));
	n++;
 	}
 anicnt = n;
 *count = n;
 }
/*
void create_fract_array(sphere *a, int *count)
{
 srand(5);
 defaultflags = GRAVITY|AIR|COLLIDEABLE;
 set_params(a+0, 400, 222, 360, 10, 0, 0, 0);
 set_params(a+1, 350, 222, 360, 10, 0, 0, 0);
 set_params(a+2, 375, 222, 390, 10, 0, 0, 0);
 set_params(a+3, 375, 222, 330, 10, 0, 0, 0);

 set_params(a+4, 375, 272, 361, 30, 0, 0, 0);
 a[0].mass = 1;
 a[1].mass = 1;
 a[2].mass = 1;
 a[3].mass = 1;
 a[4].mass = 4;
 *count = 5;
}
*/
/******************************************************
 * Sort the spheres by distance, in descending order  *
 * using merge sort, which has a guaranteed n*log n   *
 * runtime.                                           *
 *                                                    *
 * Anything simpler could have been used, as it would *
 * not lead to any performance impact, but will do    *
 * something more noticeable at larger sphere counts  *
 ******************************************************/

void sort(int l, int r)
{int i, j, m, tp=0;
 /* stage 1 */
 if (l==r) return;
 if (r-l==1) {
 	if (a[l].dist < a[r].dist) {
 		// swap the speres a[l] & a[r]
 		t[0] = a[l];
		a[l] = a[r];
		a[r] = t[0];
		}
 	return;
	}
 /* stage 2 */
 m = (l+r)/2;
 sort(l, m); sort(m+1, r);
 /* stage 3 */
 i = l; j = m+1;
 while (i<=m || j<=r) {
 	if (i>m) t[tp++]=a[j++];
		else if (j>r) t[tp++]=a[i++];
			else if (a[i].dist > a[j].dist)
				t[tp++]=a[i++];
				else t[tp++]=a[j++];
 	}
 for (i=l;i<=r;i++)
 	a[i] = t[i-l];
}

int needs_sorting(sphere *a, int count) //checks if the sphere array isn't already sorted
{
 int i;
 for (i=0;i<count-1;i++)
 	if (a[i].dist < a[i+1].dist) return 1;
 return 0;
}

void sort_fract_array(sphere *sphere_array, int count)
{
 if (count<2 || !needs_sorting(sphere_array, count)) return;
 a = sphere_array;
 sort(0, count-1);
}

void calculate_XYZ(sphere *a, double time, double *o)
{double alpha, beta, speed;
 int num = a->AniIndex;

 alpha = AS[num].alpha;
 beta = AS[num].beta;
 speed = AS[num].speed;
 //time /= 1000.0;
 o[0] = AS[num].cx + AS[num].radius*(sin(alpha)*cos(time*speed) + cos(alpha) * sin(time*speed) * sin(beta));
 o[1] = AS[num].cy + AS[num].radius*(sin(time*speed) * cos(beta));
 o[2] = AS[num].cz + AS[num].radius*(cos(alpha)*cos(time*speed) - sin(alpha) * sin(time*speed) * sin(beta));
}

void print_sphere_info(FILE *f, sphere *a)
{
    fprintf(f, " ->(x , y , z ) = (%.4lf, %.4lf, %.4lf)\n", a->x, a->y, a->z);
    fprintf(f, " ->(xi, yi, zi) = (%.4lf, %.4lf, %.4lf)\n", a->xi, a->yi, a->zi);
    fprintf(f, " ->flags = 0x%x\n", a->flags);
    fprintf(f, " ->radius = %.4lf\n", a->d);
    fprintf(f, " ->mass = %.3lf\n", a->mass);
    fprintf(f, " ->time = %.4lf\n", a->time);
    fflush(stdout);
}

void sphere_close(sphere *a, int c)
{int i;
 for (i=0;i<c;i++) {
 	if (a[i].flags & MAPPED) {
		FreeRaw(a[i].tex);
		free(a[i].tex);
		}
 	}
}
