/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

// contains some of fract.cpp's source code (merely all that is in assembly)

float w2_22 = 25.0;
float w2_shl16 = 65536.0;


// fast assembly lightformulae, returns float
static inline float lform1f( int x, int y, int lx, int ly)
{float w2_result;
 __asm __volatile(
 "	fild	%1	\n"
 "	fisub	%2	\n"
 "	fld	%%st	\n"
 "	fmulp		\n"
 "	fild	%3	\n"
 "	fisub	%4	\n"
 "	fld	%%st	\n"
 "	fmulp		\n"
 "	faddp		\n"
 "	fsqrt		\n"
 "	fsqrt		\n"
 "	fdivr	%5	\n"
 "	fstp	%0	\n"
 :"=m"(w2_result)
 :"m"(x), "m"(lx), "m"(y), "m"(ly), "m"(w2_22)
 :"memory"
 );
 return w2_result;
}


// fast assembly lightformulae, returns 16.16 fixedpoint int.
static inline int lform1(int x, int y, int lx, int ly)
{int w2_result;
 __asm __volatile(
 "	fild	%1	\n"
 "	fisub	%2	\n"
 "	fld	%%st	\n"
 "	fmulp		\n"
 "	fild	%3	\n"
 "	fisub	%4	\n"
 "	fld	%%st	\n"
 "	fmulp		\n"
 "	faddp		\n"
 "	fsqrt		\n"
 "	fsqrt		\n"
 "	fdivr	%5	\n"
 "	fmul	%6	\n"
 "	fistpl	%0	\n"
 :"=m"(w2_result)
 :"m"(x), "m"(lx), "m"(y), "m"(ly), "m"(w2_22), "m"(w2_shl16)
 :"memory"
 );
 return w2_result;
}

float w2_22_16=w2_22*65536.0;
float w2_22_8 =w2_22*256.0;


// this is the SSE version of the lighting formulae :)
// returns 16.16 fixedpoint int.
static inline int lform2(int x, int y, int lx, int ly)
{int w2_result;
 __asm __volatile(
 "	movl	%1,	%%eax	\n"
 "	subl	%2,	%%eax	\n"
 "	movl	%%eax,	%%ebx	\n"
 "	imul	%%ebx		\n"
 "	mov	%%eax,	%%ecx	\n"

 "	movl	%3,	%%eax	\n"
 "	subl	%4,	%%eax	\n"
 "	movl	%%eax,	%%ebx	\n"
 "	imul	%%ebx		\n"
 "	add	%%ecx,	%%eax	\n"

 "	cvtsi2ss %%eax,	%%xmm0	\n"
 "	movss	%5,	%%xmm1	\n"
 "	sqrtss	%%xmm0,	%%xmm0	\n"
 "	rsqrtss	%%xmm0,	%%xmm0	\n"
 "	mulss	%%xmm1,	%%xmm0	\n"
 "	cvtss2si %%xmm0, %%eax	\n"
 "	mov	%%eax,	%0	\n"

 :"=m"(w2_result)
 :"m"(x), "m"(lx), "m"(y), "m"(ly), "m"(w2_22_16)
 :"memory", "eax", "ebx", "ecx", "edx"
 );
 return w2_result;
}

// this is the integrated version of multiplycolor and lform2 - does the whole thing in a single routine
// input:  raw color from the texture. The x and y coord of the destination (where the ray hit the floor/ceiling).
//         The Lx, Ly coordinates of the light source
// output: the calculated color that should be printed to the screen
// The routine requires SSE instruction set. Supported CPUs: Pentium III and above (Intel); Athlon XP/Duron Morgan or above (AMD)
// This one uses some ideas and optimizations by barnie.

static inline Uint32	igrated(Uint32 color, int x, int y, int lx, int ly)
{
 __asm __volatile(
  // perform the light degradation math -> color multiplyer = const * ((x-lx)^2+(y-ly)^2)^-0.25
 "	movl	%0,	%%eax	\n"
 "	movl	%2,	%%ecx	\n"
 "	subl	%1,	%%eax	\n"
 "	subl	%3,	%%ecx	\n"

 "	movss	%4,	%%xmm2	\n" // move the constant to xmm2
 "	cvtsi2ss %%eax,	%%xmm0	\n" // ...try to exploit parallelism
 "	cvtsi2ss %%ecx,	%%xmm1	\n"
 "	mulss	%%xmm0,	%%xmm0	\n"
 "	mulss	%%xmm1,	%%xmm1	\n"
 "	addps	%%xmm1,	%%xmm0	\n" // xmm0 = dist^2

// the following is the barney's idea of obtaining  const / dist^0.5
 "	rsqrtss	%%xmm0,	%%xmm0	\n" // xmm0 = 1/dist
 "	mulss	%%xmm0,	%%xmm2	\n" // xmm2 = const / dist
 "	rsqrtss	%%xmm0,	%%xmm0	\n" // xmm0 = sqrt(dist)
 "	mulss	%%xmm0,	%%xmm2	\n" // xmm2 = const / sqrt(dist)
 //	load color in mm0: (via mmx):

 // note: we switch to MMX here:
 "	pxor	%%mm0,	%%mm0	\n"
 "	movd	%5,	%%mm1	\n"
 "	punpcklbw	%%mm1,	%%mm0	\n"

 "	shufps	$0,	%%xmm2,	%%xmm2	\n"
 "	cvtps2pi	%%xmm2, %%mm2	\n"
 "	pxor 	%%mm4,	%%mm4	\n"
 "	movq	%%mm2,	%%mm3	\n"
 "	packssdw	%%mm3,	%%mm2	\n"

 "	pmulhuw	%%mm2,	%%mm0	\n"

 "	packuswb	%%mm4,	%%mm0	\n"
 "	movd	%%mm0,	%%eax	\n"
 "	emms	\n"
 :
 :"m"(x), "m"(lx), "m"(y), "m"(ly), "m"(w2_22_8), "m"(color)
 :"memory", "eax", "ebx", "ecx", "edx"
 );
}

/* THIS VERSION IS DUE TO barney. */

/*
(
|------|------| ^
|      |      | |
| x0y0 | x1y0 | |
|      |      | |
|------|------| |  y
|      |      | |
| x0y1 | x1y1 | |
|      |      | |
|------|------| V

<------------->
       x

The Formula is identical to:
result = MultiplyColor(x0y0, (1-x)*(1-y)) + MultiplyColor(x0y0, (  x)*(1-y)) +
	 MultiplyColor(x0y1, (1-x)*(  y)) + MultiplyColor(x1y1, (  x)*(  y));
)
Requires SSE. Very fast :)
*/

Uint32 bilinea4(Uint32 x0y0, Uint32 x1y0, Uint32 x0y1, Uint32 x1y1, int x, int y)
{	Uint32 xx_result;
//    __m64 mm;
	__asm (
	" 	pxor		%%mm6,	%%mm6		\n"
	"	movd		%1,	%%mm0		\n"
	"	movl		$65535,	%%eax		\n"
	"	movl		%6,	%%edi		\n"
	"	movd		%2,	%%mm1		\n"
	"	movl		$65535,	%%ecx		\n"
	"	subl		%%edi,	%%ecx		\n"
	"	movd		%3,	%%mm2		\n"
	"	shll		$16,	%%edi		\n"
	"	movl		%5,	%%esi		\n"
	"	movd		%4,	%%mm3		\n"
	"	subl		%%esi,	%%eax		\n"
	"	shll		$16,	%%esi		\n"
	"	punpcklbw	%%mm6,	%%mm0		\n"
	"	orl		%%edi,	%%ecx		\n"
	"	orl		%%esi,	%%eax		\n"
	"	punpcklbw	%%mm6,	%%mm1		\n"
	"	movd		%%eax,	%%mm5		\n"
	"	movd		%%ecx,	%%mm7		\n"
	"	punpcklbw	%%mm6,	%%mm2		\n"
	"	punpcklbw	%%mm6,	%%mm3		\n"
	"	pshufw		$80,	%%mm7,	%%mm7	\n"
	"	pshufw		$68,	%%mm5,	%%mm5	\n"
	"	pmulhuw		%%mm5,	%%mm7		\n"
	"	pshufw		$0,	%%mm7,	%%mm4	\n"
	"	pshufw		$85,	%%mm7,	%%mm5	\n"
	"	pmulhuw		%%mm4,	%%mm0		\n"
	"	pshufw		$170,	%%mm7,	%%mm6	\n"
	"	pmulhuw		%%mm5,	%%mm1		\n"
	"	pshufw		$255,	%%mm7,	%%mm7	\n"
	"	pmulhuw		%%mm6,	%%mm2		\n"
	"	paddw		%%mm1,	%%mm0		\n"
	"	pmulhuw		%%mm7,	%%mm3		\n"
	"	paddw		%%mm3,	%%mm2		\n"
	"	paddw		%%mm2,	%%mm0		\n"
	"	packuswb	%%mm0,	%%mm0		\n"
	"	movd		%%mm0,	%0		\n"
	"	emms					\n"
	 :"=m"(xx_result)
	 :"m"(x0y0), "m"(x1y0), "m"(x0y1), "m"(x1y1), "m"(x), "m"(y)
	 :"memory", "eax", "edi", "ecx", "esi"
	);
	return xx_result;
}

/*********************************************************
 * blur_do_MMX: does the `blur' effect using MMX; should *
 * be about two times faster on PIII even more on P4     *
 * NOTE: count must be DIVISABLE by 8!                   *
 * Works by getting six points and processing them and   *
 * then gets additional two points and processes them as *
 * well.                                                 *
 *********************************************************/

void blur_do_MMX(Uint32 *bb, Uint32 *src, Uint32 *dst, int count, int kopy)
{
	int andval[2]= {0x1f1f1f1f, 0x1f1f1f1f};
	int cs = count;
	// load this useful value:
	__asm __volatile(
	"	movq	%0,	%%mm7\n"
	::"m"(*andval):"memory");
	count /= 8; // we are doing 8 pixels per cycle :)
	while (count--) {
		__asm __volatile(
		"	movq	%1,	%%mm0\n"
		"	movq	%0,	%%mm1\n"
		"	movq	8%1,	%%mm2\n"
		"	movq	8%0,	%%mm3\n"
		"	movq	16%1,	%%mm4\n"
		"	movq	16%0,	%%mm5\n"

		"	psrld	$3,	%%mm0\n"
		"	psrld	$3,	%%mm2\n"
		"	psrld	$3,	%%mm4\n"

		"	pand	%%mm7,	%%mm0\n"
		"	pand	%%mm7,	%%mm2\n"
		"	pand	%%mm7,	%%mm4\n"

		"	paddd	%%mm0,	%%mm1\n"
		"	paddd	%%mm2,	%%mm3\n"
		"	paddd	%%mm4,	%%mm5\n"

		"	movq	24%1,	%%mm0\n"
		"	movq	24%0,	%%mm2\n"
		"	psrld	$3,	%%mm0\n"
		"	pand	%%mm7,	%%mm0\n"
		"	paddd	%%mm2,	%%mm0\n"
		"	movq	%%mm1,	%0\n"
		"	movq	%%mm3,	8%0\n"
		"	movq	%%mm5,	16%0\n"
		"	movq	%%mm0,	24%0\n"

		:"=m"(*bb)
		:"m"(*src)
		:"memory");
		src+=8; bb+=8;
	}
	if (kopy) {
		bb-=(cs);
		count = cs/2;
		while (count--) {
			__asm __volatile(
			"	movq	%1,	%%mm0\n"
			"	movq	%%mm0,	%0\n"
			"	pxor	%%mm0,	%%mm0\n"
			"	movq	%%mm0,	%1\n"
			:"=m"(*dst),"=m"(*bb)
			:
			:"memory");
			dst+=2; bb+=2;
		}
	}
	__asm __volatile("emms");
}

/*********************************************************************
 *  MMX version of the merging function, which works on a single row *
 *-------------------------------------------------------------------*
 * INPUT:  `flr' is the pixel buffer from the floor engine           *
 *         `sph' is the pixel buffer from the sphere engine          *
 *         `multi' is an array of multipliers in 16.16 fixedpoint    *
 *             format. A value of 16384 for example, means, to blend *
 *             1/4 of the sphere pixel with 3/4 of the floor pixel   *
 *         `count' : how many pixels, must be divisable by 2.        *
 *********************************************************************/

void merge_rows(Uint32 *flr, Uint32 *sph, unsigned short *multi, int count)
{
	unsigned short ones[4] __attribute__ ((aligned (8))) = {0xffff, 0xffff, 0xffff, 0xffff};
	// load one useful value:
	__asm __volatile ("	movq	%0,	%%mm7\n"::"m"(*ones):"memory");
	count /= 2;
	while (count--) {
		__asm __volatile (
		"	pxor	%%mm6,	%%mm6\n"
		"	movq	%0,	%%mm0\n"
		"	movq	%1,	%%mm2\n"
		"	movd	%2,	%%mm4\n"
		"	movq	%%mm0,	%%mm1\n"
		"	movq	%%mm2,	%%mm3\n"
		"	punpcklbw %%mm6, %%mm0\n"
		"	punpckhbw %%mm6, %%mm1\n"
		"	punpcklbw %%mm6, %%mm2\n"
		"	punpckhbw %%mm6, %%mm3\n"
// mm0, mm1 - unpacked pixels from floor engine
// mm2, mm3 - unpacked pixels from sphere engine
// mm4 - packed multipliers

		"	movq	%%mm7,	%%mm6\n"
		"	pshufw	$0,	%%mm4,	%%mm5\n"
		"	psubw	%%mm5,	%%mm6\n"
		"	pmulhuw	%%mm5,	%%mm2\n"
		"	pmulhuw	%%mm6,	%%mm0\n"
		"	pshufw	$0x55,	%%mm4,	%%mm5\n"
		"	movq	%%mm7,	%%mm6\n"
		"	pmulhuw	%%mm5,	%%mm3\n"
		"	psubw	%%mm5,	%%mm6\n"
		"	pmulhuw	%%mm6,	%%mm1\n"
		"	paddusw	%%mm2,	%%mm0\n"
		"	paddusw	%%mm3,	%%mm1\n"
		"	packuswb %%mm7,	%%mm0\n"
		"	packuswb %%mm7,	%%mm1\n"
		"	movd	%%mm0,	%0\n"
		"	movd	%%mm1,	4%0\n"

		:"=m"(*flr)
		:"m"(*sph),"m"(*multi)
		);
		flr+=2;
		sph+=2;
		multi+=2;
	}
	__asm __volatile ("emms");
}
