/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/
/*
	This file contains the assembly alternatives of some of the shader effects routines...
*/

/* here comes a quick list of all functions with the needed stuff to tell gcc not to inline them */

void convolve_mmx_w_shifts_generic(Uint32 *src, Uint32 *dest, int resx, int resy,
					ConvolveMatrix *M, int shift) __attribute__ ((noinline));
void convolve_mmx_w_shifts_3(Uint32 *src, Uint32 *dest, int resx, int resy,
					ConvolveMatrix *M, int shift) __attribute__ ((noinline));
void convolve_mmx_w_shifts_5(Uint32 *src, Uint32 *dest, int resx, int resy,
					ConvolveMatrix *M, int shift) __attribute__ ((noinline));
void FFT_1D_complex_sse(int dir, int m, float *x, float *y) __attribute__ ((noinline));

void shader_gamma_shl(Uint32 *src, Uint32 *dest, int resx, int resy, int shift) __attribute__ ((noinline));
void shader_gamma_shr(Uint32 *src, Uint32 *dest, int resx, int resy, int shift) __attribute__ ((noinline));
void float_copy_ij_i(float *a, float *b, complex c[], int fft_size) __attribute__ ((noinline));
void float_copy_i_ij(float *a, float *b, complex c[], int fft_size) __attribute__ ((noinline));
void float_copy_ij_j(float *a, float *b, complex c[], int fft_size) __attribute__ ((noinline));
void float_copy_j_ij(float *a, float *b, complex c[], int fft_size) __attribute__ ((noinline));


#ifndef __SHADER_ASM__
#define __SHADER_ASM__

/*
	convolve_mmx_w_shifts_generic

		DESCRIPTION:
		Does the convolution effect with the given coefficient table. The maximum size of the matrix is 8x8
		(so actually useful matrix sizes are 3x3, 5x5 and 7x7, but any numbers may be used, too).
		This version is somewhat `generic'. Almost no loop unrolling is done and thus no parallelism is exploited.
		If the matrix size is 3x3 or 5x5 you may wish to use the specially coded counterparts
		`convolve_mmx_w_shifts_3' and `convolve_mmx_w_shifts_5' which are guaranteed to be at least twice as fast.

		INPUT:
		*src - Input array of RGB values
		*dest- Output framebuffer (note that this function can't do the conversion in-place, src must be different from dest)
		resx - width of the framebuffer
		resy - height of the framebuffer
		*M   - the input convoluion matrix
		shift- if the sum of the coefficients is above 1, the convolution result must be `normalized', i.e.
		       divided by that sum. It works best when the sum is a power of two and so shifts can be used
		       instead of divides. The `shift' value here means how many bits to shift-right after addition.
		       If the shift is less than 1, no actual shifting occurs.

		OUTPUT: none (*dest contains the convolution effect result)

*/
void convolve_mmx_w_shifts_generic(Uint32 *src, Uint32 *dest, int resx, int resy, ConvolveMatrix *M, int shift)
{
	int toadd, n, cols, rows;
	Sint16 __attribute__ ((aligned(64))) mm[MA3X_MAX_SIZE*MA3X_MAX_SIZE][4];
	Uint16 __attribute__ ((aligned(16))) fourones[4] = {256, 256, 256, 256};

	int i, j, k;
	for (i=0;i<M->n;i++)
		for (j=0;j<M->n;j++)
			for (k=0;k<4;k++) mm[i*M->n+j][k] = M->coeff[i][j];

	n = M->n;

	if (shift<1) shift = 0;
	toadd = (resx - M->n)*4;
	rows = resy - (n & ~1);
	cols = resx - (n & ~1);
	dest+=(resx*(M->n/2) + M->n/2);

	__asm __volatile(
	// initialize some mmx regs...
	"	pxor	%%mm7,	%%mm7\n"
	"	movd	%5, 	%%mm6\n"
	"	movd	%6,	%%mm5\n"
	// init the stuff

	// bx = M->coeff; si = src[...]; di = dst[...]
	"	movl	%2,	%%ebx\n"
	"	movl	%1,	%%esi\n"
	"	movl	%0,	%%edi\n"

	".rowloop:	\n"
	"	movl	%8,	%%edx\n"

	".balign 16	\n"
	".pixelloop:	\n"

	"	movl	%4,	%%ecx\n"
	"	pxor	%%mm0,	%%mm0\n"
	"	pxor	%%mm2,	%%mm2\n"

	".balign 16	\n"
	".y:	\n"

	"	movl	%4, 	%%eax\n"

	".balign 16	\n"
	".x:	\n"

	"	cmp	$2,	%%eax\n"
	"	jnb	.enough\n"

	"	movd	(%%esi),	%%mm1\n"
	"	punpcklbw	%%mm7,	%%mm1\n"
	"	pmullw (%%ebx),	%%mm1\n"
	"	paddsw	%%mm1,	%%mm0\n"

	"	add	$4,	%%esi\n"
	"	add	$8,	%%ebx\n"

	"	jmp	.xend\n"

	".balign 16	\n"
	".enough:	\n"
	"	movd	(%%esi),	%%mm1\n"
	"	movd	4(%%esi),	%%mm3\n"
	"	punpcklbw	%%mm7,	%%mm1\n"
	"	punpcklbw	%%mm7,	%%mm3\n"
	"	pmullw (%%ebx),	%%mm1\n"
	"	pmullw 8(%%ebx),	%%mm3\n"
	"	paddsw	%%mm1,	%%mm0\n"
	"	paddsw	%%mm3,	%%mm2\n"

	"	add	$8,	%%esi\n"
	"	add	$16,	%%ebx\n"
	"	sub	$2,	%%eax\n"
	"	jnz	.x\n"

	".balign 16	\n"
	".xend:\n"

	"	add	%3,	%%esi\n"

	"	dec	%%ecx\n"
	"	jnz	.y\n"

	"	paddsw	%%mm2,	%%mm0\n"
	"	paddsw	%%mm5,	%%mm0\n"
	"	psubusw	%%mm5,	%%mm0\n"
	"	psraw	%%mm6,	%%mm0\n"
	"	packuswb %%mm7,	%%mm0\n"
	"	movd	%%mm0,	(%%edi)\n"

	"	movl	%2,	%%ebx\n"

	"	movl	%4,	%%eax\n"
	"	xchg	%%edx,	%%ecx\n"
	"	mull	%3\n"
	"	sub	%%eax,	%%esi\n"
	"	movl	%4,	%%eax\n"
	"	mull	%4\n"
	"	shl	$2,	%%eax\n"
	"	sub	%%eax,	%%esi\n"

	"	xchg	%%edx,	%%ecx\n"

	"	add	$4,	%%edi\n"
	"	add	$4, 	%%esi\n"

	"	dec	%%edx\n"
	"	jnz	.pixelloop\n"

	"	movl	%4,	%%eax\n"
	"	andl	$0xfffffffe,	%%eax\n"
	"	shl	$2,	%%eax\n"
	"	add	%%eax,	%%esi\n"
	"	add	%%eax,	%%edi\n"

	"	decl	%7\n"
	"	jnz	.rowloop\n"

	:"=m"(dest)
	:"m"(src), "m"(mm), "m"(toadd), "m"(n), "m"(shift), "m"(*fourones), "m"(rows), "m"(cols)
	:"memory", "eax", "ebx", "ecx", "edx", "esi", "edi"
	);

	__asm __volatile("emms"); // empty multimedia state
}

/*
	convolve_mmx_w_shifts_3

		DESCRIPTION:
		Does the convolution effect with the given coefficient table. This routine uses a hardcoded assumption, that
		the matrix size is 3 (and thus it is above twice as fast than the `generic version). If the convolution
		matrix is of any other size, error occurs.

		This routine maximizes parallelism utilization.

		INPUT:
		*src - Input array of RGB values
		*dest- Output framebuffer (note that this function can't do the conversion in-place, src must be different from dest)
		resx - width of the framebuffer
		resy - height of the framebuffer
		*M   - the input convoluion matrix
		shift- if the sum of the coefficients is above 1, the convolution result must be `normalized', i.e.
		       divided by that sum. It works best when the sum is a power of two and so shifts can be used
		       instead of divides. The `shift' value here means how many bits to shift-right after addition.
		       If the shift is less than 1, no actual shifting occurs.

		OUTPUT: none (*dest contains the convolution effect result)
*/

void convolve_mmx_w_shifts_3(Uint32 *src, Uint32 *dest, int resx, int resy, ConvolveMatrix *M, int shift)
{
	int toadd, cols, rows, n=3;
	Sint16 __attribute__ ((aligned(64))) mm[9][4];
	Uint16 __attribute__ ((aligned(16))) fourones[4] = {256, 256, 256, 256};

	if (M->n!=3) {
		printf("convolve_mmx_w_shifts_3: expecting matrix size of 3, but size=%d received!\n", M->n);
		return;
	}

	int i, j, k;
	for (i=0;i<3;i++)
		for (j=0;j<3;j++)
			for (k=0;k<4;k++) mm[i*3+j][k] = M->coeff[i][j];

	if (shift<1) shift = 0;
	toadd = resx*4;
	rows = resy - 2;
	cols = resx - 2;
	dest+=(resx*1 + 1);

	__asm __volatile(
	// initialize some mmx regs...
	"	pxor	%%mm7,	%%mm7\n"
	"	movd	%5, 	%%mm6\n"
	// init the stuff

	// bx = M->coeff; si = src[...]; di = dst[...]
	"	movl	%2,	%%ebx\n"
	"	movl	%1,	%%esi\n"
	"	movl	%0,	%%edi\n"

	".rowloop_3:	\n"
	"	movl	%8,	%%edx\n"

	".balign 16	\n"
	".pixelloop_3:	\n"

	"	movd	(%%esi),	%%mm0\n"
	"	movd	4(%%esi),	%%mm1\n"
	"	movd	8(%%esi),	%%mm2\n"
	"	addl	%3,	%%esi\n"

	"	punpcklbw	%%mm7,	%%mm0\n"
	"	punpcklbw	%%mm7,	%%mm1\n"
	"	punpcklbw	%%mm7,	%%mm2\n"

	"	pmullw	(%%ebx),	%%mm0\n"
	"	pmullw	8(%%ebx),	%%mm1\n"
	"	pmullw	16(%%ebx),	%%mm2\n"

	"	movd	(%%esi),	%%mm3\n"
	"	movd	4(%%esi),	%%mm4\n"
	"	movd	8(%%esi),	%%mm5\n"
	"	addl	%3,	%%esi\n"

	"	punpcklbw	%%mm7,	%%mm3\n"
	"	punpcklbw	%%mm7,	%%mm4\n"
	"	punpcklbw	%%mm7,	%%mm5\n"

	"	pmullw	24(%%ebx),	%%mm3\n"
	"	pmullw	32(%%ebx),	%%mm4\n"
	"	pmullw	40(%%ebx),	%%mm5\n"

	"	paddsw	%%mm0,	%%mm3\n"
	"	paddsw	%%mm1,	%%mm4\n"
	"	paddsw	%%mm2,	%%mm5\n"

	"	movd	(%%esi),	%%mm0\n"
	"	movd	4(%%esi),	%%mm1\n"
	"	movd	8(%%esi),	%%mm2\n"

	"	punpcklbw	%%mm7,	%%mm0\n"
	"	punpcklbw	%%mm7,	%%mm1\n"
	"	punpcklbw	%%mm7,	%%mm2\n"

	"	pmullw	48(%%ebx),	%%mm0\n"
	"	pmullw	56(%%ebx),	%%mm1\n"
	"	pmullw	64(%%ebx),	%%mm2\n"


	"	paddsw	%%mm3,	%%mm0\n"
	"	paddsw	%%mm4,	%%mm1\n"
	"	paddsw	%%mm5,	%%mm2\n"

	"	paddsw	%%mm1,	%%mm0\n"
	"	paddsw	%6,	%%mm3\n"
	"	paddsw	%%mm3,	%%mm0\n"

	"	psubusw	%6,	%%mm0\n"
	"	psraw	%%mm6,	%%mm0\n"
	"	packuswb %%mm7,	%%mm0\n"
	"	movd	%%mm0,	(%%edi)\n"

	"	subl	%3,	%%esi\n"
	"	subl	%3,	%%esi\n"

	"	add	$4,	%%edi\n"
	"	add	$4, 	%%esi\n"

	"	dec	%%edx\n"
	"	jnz	.pixelloop_3\n"

	"	movl	%4,	%%eax\n"
	"	andl	$0xfffffffe,	%%eax\n"
	"	shl	$2,	%%eax\n"
	"	add	%%eax,	%%esi\n"
	"	add	%%eax,	%%edi\n"

	"	decl	%7\n"
	"	jnz	.rowloop_3\n"

	:"=m"(dest)
	:"m"(src), "m"(mm), "m"(toadd), "m"(n), "m"(shift), "m"(*fourones), "m"(rows), "m"(cols)
	:"memory", "eax", "ebx", "ecx", "edx", "esi", "edi"
	);

	__asm __volatile("emms"); // empty multimedia state
}

// this function assumes that M->n is 5.
/*
	convolve_mmx_w_shifts_5

		DESCRIPTION: see the info for convolve_mmx_w_shifts_3, this function does the same,
	but is hardcoded and loop-unrolled for matrix size of 5, attempting to maximize parallelism utilization
*/
void convolve_mmx_w_shifts_5(Uint32 *src, Uint32 *dest, int resx, int resy, ConvolveMatrix *M, int shift)
{
	int toadd, cols, rows, n=5;
	Sint16 __attribute__ ((aligned(64))) mm[25][4];
	Uint16 __attribute__ ((aligned(16))) fourones[4] = {256, 256, 256, 256};

	if (M->n!=5) {
		printf("convolve_mmx_w_shifts_5: expecting matrix size of 5, but size=%d received!\n", M->n);
		return;
	}

	int i, j, k;
	for (i=0;i<5;i++)
		for (j=0;j<5;j++)
			for (k=0;k<4;k++) mm[i*5+j][k] = M->coeff[i][j];

	if (shift<1) shift = 0;
	toadd = resx*4;
	rows = resy - 4;
	cols = resx - 4;
	dest+=(resx*2 + 2);

	__asm __volatile(
	// initialize some mmx regs...
	"	pxor	%%mm7,	%%mm7\n"
//	"	movd	%5, 	%%mm6\n"
	// init the stuff

	// bx = M->coeff; si = src[...]; di = dst[...]
	"	movl	%2,	%%ebx\n"
	"	movl	%1,	%%esi\n"
	"	movl	%0,	%%edi\n"

	".rowloop_5:	\n"
	"	movl	%8,	%%edx\n"

	".balign 16	\n"
	".pixelloop_5:	\n"

	// row 1:
	"	movd	  (%%esi),	%%mm2\n"
	"	movd	 4(%%esi),	%%mm3\n"
	"	movd	 8(%%esi),	%%mm4\n"
	"	movd	12(%%esi),	%%mm5\n"
	"	movd	16(%%esi),	%%mm6\n"
	"	addl	%3,	%%esi\n"

	"	punpcklbw	%%mm7,	%%mm2\n"
	"	punpcklbw	%%mm7,	%%mm3\n"
	"	punpcklbw	%%mm7,	%%mm4\n"
	"	punpcklbw	%%mm7,	%%mm5\n"
	"	punpcklbw	%%mm7,	%%mm6\n"

	"	pmullw	  (%%ebx),	%%mm2\n"
	"	pmullw	 8(%%ebx),	%%mm3\n"
	"	pmullw	16(%%ebx),	%%mm4\n"
	"	pmullw	24(%%ebx),	%%mm5\n"
	"	pmullw	32(%%ebx),	%%mm6\n"

	"	paddsw	%%mm2,	%%mm4\n"
	"	paddsw	%%mm3,	%%mm5\n"

	// row 2:
	"	movd	  (%%esi),	%%mm0\n"
	"	movd	 4(%%esi),	%%mm1\n"

	"	paddsw	%%mm4,	%%mm6\n"

	"	movd	 8(%%esi),	%%mm2\n"
	"	movd	12(%%esi),	%%mm3\n"
	"	movd	16(%%esi),	%%mm4\n"
	"	addl	%3,	%%esi\n"

	"	punpcklbw	%%mm7,	%%mm0\n"
	"	punpcklbw	%%mm7,	%%mm1\n"
	"	punpcklbw	%%mm7,	%%mm2\n"
	"	punpcklbw	%%mm7,	%%mm3\n"
	"	punpcklbw	%%mm7,	%%mm4\n"

	"	pmullw	 40(%%ebx),	%%mm0\n"
	"	pmullw	 48(%%ebx),	%%mm1\n"
	"	pmullw	 56(%%ebx),	%%mm2\n"
	"	pmullw	 64(%%ebx),	%%mm3\n"
	"	pmullw	 72(%%ebx),	%%mm4\n"

	"	paddsw	%%mm0,	%%mm2\n"
	"	paddsw	%%mm1,	%%mm3\n"
	"	paddsw	%%mm4,	%%mm5\n"
	"	paddsw	%%mm2,	%%mm6\n"

	// row 3:
	"	movd	  (%%esi),	%%mm0\n"
	"	movd	 4(%%esi),	%%mm1\n"

	"	paddsw	%%mm3,	%%mm5\n"

	"	movd	 8(%%esi),	%%mm2\n"
	"	movd	12(%%esi),	%%mm3\n"
	"	movd	16(%%esi),	%%mm4\n"
	"	addl	%3,	%%esi\n"

	"	punpcklbw	%%mm7,	%%mm0\n"
	"	punpcklbw	%%mm7,	%%mm1\n"
	"	punpcklbw	%%mm7,	%%mm2\n"
	"	punpcklbw	%%mm7,	%%mm3\n"
	"	punpcklbw	%%mm7,	%%mm4\n"

	"	pmullw	 80(%%ebx),	%%mm0\n"
	"	pmullw	 88(%%ebx),	%%mm1\n"
	"	pmullw	 96(%%ebx),	%%mm2\n"
	"	pmullw	104(%%ebx),	%%mm3\n"
	"	pmullw	112(%%ebx),	%%mm4\n"

	"	paddsw	%%mm0,	%%mm2\n"
	"	paddsw	%%mm1,	%%mm3\n"
	"	paddsw	%%mm4,	%%mm5\n"
	"	paddsw	%%mm2,	%%mm6\n"

	// row 4:
	"	movd	  (%%esi),	%%mm0\n"
	"	movd	 4(%%esi),	%%mm1\n"

	"	paddsw	%%mm3,	%%mm5\n"

	"	movd	 8(%%esi),	%%mm2\n"
	"	movd	12(%%esi),	%%mm3\n"
	"	movd	16(%%esi),	%%mm4\n"
	"	addl	%3,	%%esi\n"

	"	punpcklbw	%%mm7,	%%mm0\n"
	"	punpcklbw	%%mm7,	%%mm1\n"
	"	punpcklbw	%%mm7,	%%mm2\n"
	"	punpcklbw	%%mm7,	%%mm3\n"
	"	punpcklbw	%%mm7,	%%mm4\n"

	"	pmullw	120(%%ebx),	%%mm0\n"
	"	pmullw	128(%%ebx),	%%mm1\n"
	"	pmullw	136(%%ebx),	%%mm2\n"
	"	pmullw	144(%%ebx),	%%mm3\n"
	"	pmullw	152(%%ebx),	%%mm4\n"

	"	paddsw	%%mm0,	%%mm2\n"
	"	paddsw	%%mm1,	%%mm3\n"
	"	paddsw	%%mm4,	%%mm5\n"
	"	paddsw	%%mm2,	%%mm6\n"

	// row 5:
	"	movd	  (%%esi),	%%mm0\n"
	"	movd	 4(%%esi),	%%mm1\n"

	"	paddsw	%%mm3,	%%mm5\n"

	"	movd	 8(%%esi),	%%mm2\n"
	"	movd	12(%%esi),	%%mm3\n"
	"	movd	16(%%esi),	%%mm4\n"

	"	punpcklbw	%%mm7,	%%mm0\n"
	"	punpcklbw	%%mm7,	%%mm1\n"
	"	punpcklbw	%%mm7,	%%mm2\n"
	"	punpcklbw	%%mm7,	%%mm3\n"
	"	punpcklbw	%%mm7,	%%mm4\n"

	"	pmullw	160(%%ebx),	%%mm0\n"
	"	pmullw	168(%%ebx),	%%mm1\n"
	"	pmullw	176(%%ebx),	%%mm2\n"
	"	pmullw	184(%%ebx),	%%mm3\n"
	"	pmullw	192(%%ebx),	%%mm4\n"

// situation here: we have data all over the first 7 registers.. we must first sum it up...

	"	paddsw	%%mm5,	%%mm6\n"
	"	paddsw	%%mm0,	%%mm1\n"
	"	paddsw	%%mm2,	%%mm3\n"
	"	paddsw	%%mm4,	%%mm6\n"
	"	paddsw	%6,	%%mm1\n"
	"	movd	%5,	%%mm0\n"

	"	paddsw	%%mm6,	%%mm3\n"
	"	paddsw	%%mm3,	%%mm1\n"


	"	psubusw	%6,	%%mm1\n"
	"	psraw	%%mm0,	%%mm1\n"
	"	packuswb %%mm7,	%%mm1\n"
	"	movd	%%mm1,	(%%edi)\n"

	"	movl	%3,	%%eax\n"
	"	shl	$2,	%%eax\n"
	"	subl	%%eax,	%%esi\n"

	"	add	$4,	%%edi\n"
	"	add	$4, 	%%esi\n"

	"	dec	%%edx\n"
	"	jnz	.pixelloop_5\n"

	"	movl	%4,	%%eax\n"
	"	andl	$0xfffffffe,	%%eax\n"
	"	shl	$2,	%%eax\n"
	"	add	%%eax,	%%esi\n"
	"	add	%%eax,	%%edi\n"

	"	decl	%7\n"
	"	jnz	.rowloop_5\n"

	:"=m"(dest)
	:"m"(src), "m"(mm), "m"(toadd), "m"(n), "m"(shift), "m"(*fourones), "m"(rows), "m"(cols)
	:"memory", "eax", "ebx", "ecx", "edx", "esi", "edi"
	);

	__asm __volatile("emms"); // empty multimedia state
}

/*
	convolve_mmx_w_shifts:

		DESCRIPTION: does the convolution effect on a given coefficien matrix (up to 8x8) with sum normalization
	using bitwize right shifts (note that this can only be done when the sum of the coeficients of the convolution matrix
	is either zero, or a power of two).
		Runs very fast in the special case of 3x3 or 5x5 matrices, as a special hardcoded and tightly optimized
	(loop unrolled) versions of the effect are called.
		All functions use the "original" MMX instruction set (it is OK on PentiumII and MMX-enabled Pentium-Is)

		INPUT:
		*src - Input array of RGB values
		*dest- Output framebuffer (note that this function can't do the conversion in-place, src must be different from dest)
		resx - width of the framebuffer
		resy - height of the framebuffer
		*M   - the input convoluion matrix
		shift- if the sum of the coefficients is above 1, the convolution result must be `normalized', i.e.
		       divided by that sum. It works best when the sum is a power of two and so shifts can be used
		       instead of divides. The `shift' value here means how many bits to shift-right after addition.
		       If the shift is less than 1, no actual shifting occurs.

		OUTPUT: none (*dest contains the convolution effect result)

*/
static inline void convolve_mmx_w_shifts(Uint32 *src, Uint32 *dest, int resx, int resy, ConvolveMatrix *M, int shift)
{
	//printf("called convolve_mmx_w_shifts(...,...,....,...,...,%d)\n", shift);
	switch (M->n) { // if we have a match for that matrix size, we should use it - it would be the fastest function
		case 3:  convolve_mmx_w_shifts_3(src, dest, resx, resy, M, shift); break;
		case 5:  convolve_mmx_w_shifts_5(src, dest, resx, resy, M, shift); break;
		default: convolve_mmx_w_shifts_generic(src, dest, resx, resy, M, shift);
	}
}

// shifts left all pixel data by the specified shift. This is the same as raising the gamma by 2^shift
// NOTE: uses MMX and requires the number of pixels % 4 == 0
void shader_gamma_shl(Uint32 *src, Uint32 *dest, int resx, int resy, int shift)
{
	int count = (resx * resy) / 4;

	// the innermost loop is 4 times unrolled
	__asm __volatile(
// prepare the loop:
	//"	ulaw\n"
	"	pxor	%%mm7,	%%mm7\n"
	"	movd	%2,	%%mm6\n"
	"	movl	%3,	%%eax\n"
	"	movl	%0,	%%edi\n"
	"	movl	%1,	%%esi\n"

	".balign 16\n"
	".shader_gamma_shl_loop:"

	"	movd	(%%esi),	%%mm0\n"
	"	movd	4(%%esi),	%%mm1\n"
	"	movd	8(%%esi),	%%mm2\n"
	"	movd	12(%%esi),	%%mm3\n"

	"	punpcklbw	%%mm7,	%%mm0\n"
	"	punpcklbw	%%mm7,	%%mm1\n"
	"	punpcklbw	%%mm7,	%%mm2\n"
	"	punpcklbw	%%mm7,	%%mm3\n"

	"	psllw	%%mm6,	%%mm0\n"
	"	psllw	%%mm6,	%%mm1\n"
	"	psllw	%%mm6,	%%mm2\n"
	"	psllw	%%mm6,	%%mm3\n"

	"	packuswb	%%mm7,	%%mm0\n"
	"	packuswb	%%mm7,	%%mm1\n"
	"	packuswb	%%mm7,	%%mm2\n"
	"	packuswb	%%mm7,	%%mm3\n"

	"	movd	%%mm0,	(%%edi)\n"
	"	movd	%%mm1,	4(%%edi)\n"
	"	movd	%%mm2,	8(%%edi)\n"
	"	movd	%%mm3,	12(%%edi)\n"

	"	add	$16,	%%esi\n"
	"	add	$16,	%%edi\n"
	"	dec	%%eax\n"
	"	jnz	.shader_gamma_shl_loop\n"
	"	emms\n"
	:"=m"(dest)
	:"m"(src), "m"(shift), "m"(count)
	:"memory", "eax", "esi", "edi");
}

// shifts right all pixel data by the specified shift. This is the same as lowering the gamma by 2^shift
// NOTE: uses MMX and requires the number of pixels % 4 == 0
void shader_gamma_shr(Uint32 *src, Uint32 *dest, int resx, int resy, int shift)
{
	int count = (resx * resy) / 4;

	// the innermost loop is 4 times unrolled
	__asm __volatile(
// prepare the loop:
	//"	ulaw\n"
	"	pxor	%%mm7,	%%mm7\n"
	"	movd	%2,	%%mm6\n"
	"	movl	%3,	%%eax\n"
	"	movl	%0,	%%edi\n"
	"	movl	%1,	%%esi\n"

	".balign 16\n"
	".shader_gamma_shr_loop:"

	"	movd	(%%esi),	%%mm0\n"
	"	movd	4(%%esi),	%%mm1\n"
	"	movd	8(%%esi),	%%mm2\n"
	"	movd	12(%%esi),	%%mm3\n"

	"	punpcklbw	%%mm7,	%%mm0\n"
	"	punpcklbw	%%mm7,	%%mm1\n"
	"	punpcklbw	%%mm7,	%%mm2\n"
	"	punpcklbw	%%mm7,	%%mm3\n"

	"	psrlw	%%mm6,	%%mm0\n"
	"	psrlw	%%mm6,	%%mm1\n"
	"	psrlw	%%mm6,	%%mm2\n"
	"	psrlw	%%mm6,	%%mm3\n"

	"	packuswb	%%mm7,	%%mm0\n"
	"	packuswb	%%mm7,	%%mm1\n"
	"	packuswb	%%mm7,	%%mm2\n"
	"	packuswb	%%mm7,	%%mm3\n"

	"	movd	%%mm0,	(%%edi)\n"
	"	movd	%%mm1,	4(%%edi)\n"
	"	movd	%%mm2,	8(%%edi)\n"
	"	movd	%%mm3,	12(%%edi)\n"

	"	add	$16,	%%esi\n"
	"	add	$16,	%%edi\n"
	"	dec	%%eax\n"
	"	jnz	.shader_gamma_shr_loop\n"
	"	emms\n"
	:"=m"(dest)
	:"m"(src), "m"(shift), "m"(count)
	:"memory", "eax", "esi", "edi");
}


/*
	FFT_1D_complex_sse

		DESCRIPTION: SSE version of the 1-dimensional 2^m-point FFT, operating in the field of complex numbers.

			The code does four 2^m-point FFTs in parallel. The first FFT is performed on the multiple-of-four indicies
		of the real[] and imag[] arrays. The second FFT is on those indicies, which, divided by 4, yield 1 as a remainder, and
		so on. So if you want to make a 1024-point FFT, you must pass two 4096-element arrays, the first with real data
		and the second with imaginary data.
			There's no way to prevent making four ffts in parallel, but a simple workaround is (persume my_fft_func does
		a n-point fft:)

		void my_fft_func(complex *a, int n, int dir)
		{
			int i;
			float real[MAX_FFT_SIZE], imag[MAX_FFT_SIZE];
			for (i=0;i<n;i++) {
				real[4*i] = a[i].real;
				imag[4*i] = a[i].real;
			}
			FFT_1D_complex_sse(dir, log2(n), real, imag);
			for (i=0;i<n;i++) {
				a[i].real = real[4*i];
				a[i].imag = imag[4*i];
			}
		}

		This would be marginally slower than a dedicated scalar version of the FFT, since scalar arithmetic is not faster
		on the latest CPU's

		INPUT:
			dir - direction of the FFT 1 = forward(decimation in frequency), -1 = backward (decimation in time)
			m   - log2 of number of points
			x   - pointer to real floatingpoint input data, containing 4 * 2^m points
			y   - pointer to imaginary floatingpoint data, containing 4 * 2^m points

		OUTPUT:
			the conversion is done in-place
*/

void FFT_1D_complex_sse(int dir, int m, float *x, float *y)
{
   long i,nn,j,l;
   float __attribute__ ((aligned(16))) c1[4],c2[4];
float __attribute__ ((aligned(16))) fftconsts[20] =	{ 1.0,  1.0,  1.0,  1.0,
							  0.0,  0.0,  0.0,  0.0,
							 -1.0, -1.0, -1.0, -1.0,
							  0.5,  0.5,  0.5,  0.5,
							 -2.0, -2.0, -2.0, -2.0
							};

   /* Calculate the number of points */
   nn = 1;
   for (i=0;i<m;i++)
      nn *= 2;

   __asm __volatile(

   	"	movl	%2,	%%edx\n" //edx=i2
	"	decl	%2\n"
	"	shr	%%edx\n"
	"	xor	%%ebx,	%%ebx\n" //ebx=j
	"	movl	%0,	%%esi\n"
	"	movl	%1,	%%edi\n"
	"	xor	%%eax,	%%eax\n" //eax=i

	".balign 16\n"
	".fft_l1:\n"

	"	cmp	%%ebx,	%%eax\n"
	"	jge	.fftb1\n"

	"	shl	$4,	%%eax\n"
	"	shl	$4,	%%ebx\n"

	"	movaps	(%%esi, %%eax),	%%xmm0\n" // the fastest xchg known to man!
	"	movaps	(%%edi, %%eax),	%%xmm1\n"
	"	movaps	(%%esi, %%ebx),	%%xmm2\n"
	"	movaps	(%%edi, %%ebx),	%%xmm3\n"

	"	movaps	%%xmm0,	(%%esi, %%ebx)\n"
	"	movaps	%%xmm1,	(%%edi, %%ebx)\n"
	"	movaps	%%xmm2,	(%%esi, %%eax)\n"
	"	movaps	%%xmm3,	(%%edi, %%eax)\n"

	"	shr	$4,	%%eax\n"
	"	shr	$4,	%%ebx\n"


	".balign 16\n"
	".fftb1:\n"

	"	movl	%%edx,	%%ecx\n" // ecx=k
	".balign 16\n"
	".fftwhile1:\n"
	"	cmp	%%ebx,	%%ecx\n"
	"	jg	.fftb2\n"

	"	sub	%%ecx,	%%ebx\n"
	"	shr	%%ecx\n"
	"	jmp	.fftwhile1\n"

	".balign 16\n"
	".fftb2:\n"

	"	addl	%%ecx,	%%ebx\n"


	"	inc	%%eax\n"
	"	cmpl	%2,	%%eax\n"
	"	jb	.fft_l1\n"
	"	incl	%2\n"

   :"=m"(x), "=m"(y)
   :"m"(nn)
   :"memory", "eax", "ebx", "ecx", "edx", "esi", "edi"
   );

   /* Compute the FFT */

l = m;

__asm __volatile (// mapping: eax = i, ebx = i1, ecx = l1, edx = l2. esi = x, edi = y

//initialize:
	"	movl	%0,	%%esi\n"
	"	movl	%1,	%%edi\n"

	"	movaps	32%7,	%%xmm0\n"
	"	movaps	16%7,	%%xmm1\n"
	"	movaps	%%xmm0, %2\n" // move -1 to c1
	"	movaps	%%xmm1,	%3\n" // move  0 to c2

	"	movl	$1,	%%edx\n"

/* XMM mappings
   0, 1 - accumulators
   2, 3 - u1, u2
   4, 5 - t1, t2
*/
	".balign 16\n"
	".fft_log2:\n"

	"	mov	%%edx,	%%ecx\n"
	"	shl	%%edx\n"

	"	movaps	%7,	%%xmm2\n" // move 1 to u1
	"	movaps	16%7,	%%xmm3\n" // move 0 to u2
	"	movl	$0,	%5\n"     // j = 0

	".balign 16\n"
	".fft_c1:\n"
	"	movl	%5,	%%eax\n" // i= j

	".balign 16\n"
	".fft_c2:\n"

	"	mov	%%eax,	%%ebx\n"
	"	add	%%ecx,	%%ebx\n" // ebx = i+l1

	"	shl	$4,	%%ebx\n"
	"	shl	$4,	%%eax\n"

	"	movaps	(%%esi, %%ebx),	%%xmm0\n" // xmm0 = x[i1]
	"	movaps	(%%edi,	%%ebx),	%%xmm1\n" // xmm1 = y[i1]
	"	movaps	%%xmm0,	%%xmm4\n"
	"	movaps	%%xmm1,	%%xmm5\n"

	"	mulps	%%xmm3,	%%xmm0\n"	// xmm0 = x[i1]*u2
	"	mulps	%%xmm3,	%%xmm1\n"	// xmm1 = y[i1]*u2
	"	mulps	%%xmm2,	%%xmm4\n"	// t1 = x[i1]*u1
	"	mulps	%%xmm2,	%%xmm5\n"	// t2 = y[i1]*u1

	"	subps	%%xmm1,	%%xmm4\n"	// t1 = x[i1]*u1 - y[i1]*u2
	"	addps	%%xmm0,	%%xmm5\n"	// t2 = x[i1]*u2 + y[i1]*u1

	"	movaps	(%%esi, %%eax),	%%xmm6\n" // xmm6 = x[i]
	"	movaps	(%%edi, %%eax),	%%xmm7\n" // xmm7 = y[i]

	"	movaps	%%xmm6,	%%xmm0\n"  // clone to xmm0 and xmm1
	"	movaps	%%xmm7,	%%xmm1\n"

	"	subps	%%xmm4,	%%xmm6\n" // xmm6 = x[i] - t1
	"	subps	%%xmm5,	%%xmm7\n" // xmm7 = x[i] - t2
	"	addps	%%xmm4,	%%xmm0\n" // xmm0 = x[i] + t1
	"	addps	%%xmm5,	%%xmm1\n" // xmm1 = x[i] + t2

	"	movaps	%%xmm6,	(%%esi,	%%ebx)\n" // x[i1] = x[i] - t1
	"	movaps	%%xmm7,	(%%edi,	%%ebx)\n" // y[i1] = y[i] - t2
	"	movaps	%%xmm0,	(%%esi,	%%eax)\n" // x[i] += t1
	"	movaps	%%xmm1,	(%%edi,	%%eax)\n" // y[i] += t2

	"	shr	$4,	%%ebx\n"
	"	shr	$4,	%%eax\n"

	"	addl	%%edx,	%%eax\n"
	"	cmpl	%6,	%%eax\n"
	"	jl	.fft_c2\n"

	"	movaps	%%xmm2,	%%xmm0\n" // xmm0 == xmm2 == u1
	"	movaps	%%xmm3,	%%xmm1\n" // xmm1 == xmm3 == u2

/*
         z =  u1 * c1 - u2 * c2;
         u2 = u1 * c2 + u2 * c1;
         u1 = z; (c1 = %2, c2 = %3)
*/
	"	mulps	%2,	%%xmm2\n" // xmm2 = u1 * c1
	"	mulps	%2,	%%xmm3\n" // xmm3 = u2 * c1
	"	mulps	%3,	%%xmm0\n" // xmm0 = u1 * c2
	"	mulps	%3,	%%xmm1\n" // xmm1 = u2 * c2
	"	addps	%%xmm0,	%%xmm3\n" // new_u2 = u2 * c1 + u1 * c2
	"	subps	%%xmm1,	%%xmm2\n" // new_u1 = u1 * c1 - u2 * c2

	"	incl	%5\n"
	"	cmpl	%%ecx,	%5\n"
	"	jl	.fft_c1\n"
/*
      c2 = sqrt((1.0 - c1) * 0.5);
      if (dir == 1)
         c2 = -c2;
      c1 = sqrt((1.0 + c1) * 0.5);
*/
	"	movaps	%7,	%%xmm0\n" // xmm0 = 1.0
	"	movaps	%%xmm0,	%%xmm1\n" // xmm1 = 1.0
	"	subps	%2,	%%xmm0\n" // xmm0 = 1.0 - c1
	"	addps	%2,	%%xmm1\n" // xmm1 = 1.0 + c1
	"	mulps	48%7,	%%xmm0\n" // xmm0 = (1.0 - c1) * 0.5
	"	mulps	48%7,	%%xmm1\n" // xmm1 = (1.0 + c1) * 0.5
	"	sqrtps	%%xmm0,	%%xmm0\n" // xmm0 ^= 0.5
	"	sqrtps	%%xmm1,	%%xmm1\n" // xmm1 ^= 0.5

	"	movaps	%%xmm0,	%3\n"	// c2 = sqrt((1.0 - c1) * 0.5)
	"	movaps	%%xmm1,	%2\n"   // c1 = sqrt((1.0 + c1) * 0.5)

	"	cmpl	$1,	%8\n"
	"	jne	.fft_invb\n"

	"	mulps	64%7,	%%xmm0\n" // if (dir == 1) c2 = -c2
	"	addps	%3,	%%xmm0\n"
	"	movaps	%%xmm0,	%3\n"

	".balign 16\n"
	".fft_invb:\n"

	"	decl	%4\n" // l--
	"	jnz	.fft_log2\n" // if l ... jump

   /* Scaling for forward transform */

	"	cmpl	$1,	%8\n"
	"	jne	.fft_nscale\n"

	"	cvtsi2ss	%6,	%%xmm0\n"
	"	movaps	%%xmm0,	%%xmm1\n"
	"	shufps	$0,	%%xmm0,	%%xmm1\n" //xmm1 = nn nn nn nn

	"	movaps	%7,	%%xmm0\n" // xmm0 = 1 1 1 1
	"	divps	%%xmm1,	%%xmm0\n" // xmm0 = 1/nn 1/nn 1/nn 1/nn
	//"	addps	%%xmm0,	%%xmm0\n" // xmm0 = 2/nn 2/nn 2/nn 2/nn

	"	movl	%6,	%%ecx\n"
	".balign 16\n"
	".fft_scale:\n"

	"	movaps	%%xmm0,	%%xmm1\n"
	"	movaps	%%xmm0,	%%xmm2\n"
	"	mulps	(%%esi),	%%xmm1\n"
	"	mulps	(%%edi),	%%xmm2\n"
	"	movaps	%%xmm1,	(%%esi)\n"
	"	movaps	%%xmm2,	(%%edi)\n"

	"	addl	$16,	%%esi\n"
	"	addl	$16,	%%edi\n"

	"	decl	%%ecx\n"
	"	jnz	.fft_scale\n"

	".balign 16\n"
	".fft_nscale:\n"

	"	emms\n"

/*
	FFT Constants:
	offset		0	16	32	48	64
	constant	+1	 0	-1	0.5	-2
*/
:"=m"(x), "=m"(y), "=m"(*c1), "=m"(*c2), "=m"(l), "=m"(j)
:"m"(nn), "m"(*fftconsts) , "m"(dir)
:"memory", "eax", "ebx", "ecx", "edx", "esi", "edi"
);

}

/*
	this function replaces the following code:
      		for (i=0;i<fft_size;i++) {
         		real[4*i  ] = c[i][4*j  ].re;
         		imag[4*i  ] = c[i][4*j  ].im;
         		real[4*i+1] = c[i][4*j+1].re;
         		imag[4*i+1] = c[i][4*j+1].im;
         		real[4*i+2] = c[i][4*j+2].re;
         		imag[4*i+2] = c[i][4*j+2].im;
         		real[4*i+3] = c[i][4*j+3].re;
         		imag[4*i+3] = c[i][4*j+3].im;
      		}
*/
void float_copy_ij_i(float *a, float *b, complex c[], int fft_size)
{
	__asm __volatile(
	"	movl	%3,	%%eax\n"
	//"	movl	%%eax,	%%edx\n"
	//"	shl	$3,	%%edx\n"
	"	movl	%0,	%%esi\n"
	"	movl	%1,	%%edi\n"
	"	movl	%2,	%%ebx\n"

	".balign 16\n"
	".copy_ij_i_loop:\n"

	"	movaps	(%%ebx),	%%xmm0\n" // xmm0 = re[0] im[0] re[1] im[1]
	"	movaps	16(%%ebx),	%%xmm1\n" // xmm1 = re[2] im[2] re[3] im[3]
	"	movaps	%%xmm0,	%%xmm2\n"	  // xmm2 = re[0] im[0] re[1] im[1]
	"	movaps	%%xmm1,	%%xmm3\n"	  // xmm3 = re[2] im[2] re[3] im[3]
	"	shufps	$0x88,	%%xmm1,	%%xmm0\n" // 10 00 10 00 // xmm0 = re[0] re[1] re[2] re[3]
	"	shufps	$0xdd,	%%xmm3,	%%xmm2\n" // 11 01 11 01 // xmm2 = im[0] im[1] im[2] im[3]
	"	movaps	%%xmm0,	(%%esi)\n"
	"	movaps	%%xmm2,	(%%edi)\n"

	"	add	$8192,	%%ebx\n"
	"	add	$16,	%%esi\n"
	"	add	$16,	%%edi\n"

	"	decl	%%eax\n"
	"	jnz	.copy_ij_i_loop\n"

	"	emms\n"
	:"=m"(a), "=m"(b)
	:"m"(c), "m"(fft_size)
	:"memory", "eax", "ebx", "edx", "esi", "edi"
	);

}

/*
	this function replaces the following code:
      		for (i=0;i<fft_size;i++) {
         		 c[i][4*j  ].re = real[4*i  ];
         		 c[i][4*j  ].im = imag[4*i  ];
         		 c[i][4*j+1].re = real[4*i+1];
         		 c[i][4*j+1].im = imag[4*i+1];
         		 c[i][4*j+2].re = real[4*i+2];
         		 c[i][4*j+2].im = imag[4*i+2];
         		 c[i][4*j+3].re = real[4*i+3];
         		 c[i][4*j+3].im = imag[4*i+3];
      		}
*/
void float_copy_i_ij(float *a, float *b, complex c[], int fft_size)
{
	__asm __volatile(
	"	movl	%3,	%%eax\n"
	//"	movl	%%eax,	%%edx\n"
	//"	shl	$3,	%%edx\n"
	"	movl	%0,	%%esi\n"
	"	movl	%1,	%%edi\n"
	"	movl	%2,	%%ebx\n"

	".balign 16\n"
	".copy_i_ij_loop:\n"

	"	movaps	(%%esi),	%%xmm0\n" // xmm0 = re[0] re[1] re[2] re[3]
	"	movaps	(%%edi),	%%xmm1\n" // xmm1 = im[0] im[1] im[2] im[3]
	"	movaps	%%xmm0,	%%xmm2\n"	  // xmm2 = re[0] re[1] re[2] re[3]
	"	movaps	%%xmm1,	%%xmm3\n"	  // xmm3 = im[0] im[1] im[2] im[3]

	"	unpcklps	%%xmm1,	%%xmm0\n" // xmm0 = re[0] im[0] re[1] im[1]
	"	unpckhps	%%xmm3,	%%xmm2\n" // xmm2 = re[2] im[2] re[3] im[3]

	"	movaps	%%xmm0,	(%%ebx)\n"
	"	movaps	%%xmm2,	16(%%ebx)\n"

	"	add	$8192,	%%ebx\n"
	"	add	$16,	%%esi\n"
	"	add	$16,	%%edi\n"

	"	decl	%%eax\n"
	"	jnz	.copy_i_ij_loop\n"

	"	emms\n"
	:"=m"(a), "=m"(b)
	:"m"(c), "m"(fft_size)
	:"memory", "eax", "ebx", "edx", "esi", "edi"
	);

}


/***************************************************************************************************************************************/
/***************************************************************************************************************************************/
/***************************************************************************************************************************************/
/***************************************************************************************************************************************/
/***************************************************************************************************************************************/
/***************************************************************************************************************************************/
/***************************************************************************************************************************************/
/***************************************************************************************************************************************/
/***************************************************************************************************************************************/
/***************************************************************************************************************************************/
/***************************************************************************************************************************************/
/***************************************************************************************************************************************/


/*
	this function replaces the following code:
      		for (j=0;j<fft_size;k++) {
         		real[4*j  ] = c[4*i  ][j].re;
         		imag[4*j  ] = c[4*i  ][j].im;
         		real[4*j+1] = c[4*i+1][j].re;
         		imag[4*j+1] = c[4*i+1][j].im;
         		real[4*j+2] = c[4*i+2][j].re;
         		imag[4*j+2] = c[4*i+2][j].im;
         		real[4*j+3] = c[4*i+3][j].re;
         		imag[4*j+3] = c[4*i+3][j].im;
      		}
*/
void float_copy_ij_j(float *a, float *b, complex c[], int fft_size)
{
	__asm __volatile(
	"	movl	%3,	%%eax\n"
	"	movl	%0,	%%esi\n"
	"	movl	%1,	%%edi\n"
	"	movl	%2,	%%ebx\n"

	".balign 16\n"
	".copy_ij_j_loop:\n"

	"	movlps	(%%ebx),	%%xmm0\n"
	"	movhps	 8192(%%ebx),	%%xmm0\n" // xmm0 = re[0] im[0] re[1] im[1]
	"	movlps	16384(%%ebx),	%%xmm1\n"
	"	movhps	24576(%%ebx),	%%xmm1\n" // xmm1 = re[2] im[2] re[3] im[3]
	"	movaps	%%xmm0,	%%xmm2\n"	  // xmm2 = re[0] im[0] re[1] im[1]
	"	movaps	%%xmm1,	%%xmm3\n"	  // xmm3 = re[2] im[2] re[3] im[3]
	"	shufps	$0x88,	%%xmm1,	%%xmm0\n" // 10 00 10 00 // xmm0 = re[0] re[1] re[2] re[3]
	"	shufps	$0xdd,	%%xmm3,	%%xmm2\n" // 11 01 11 01 // xmm2 = im[0] im[1] im[2] im[3]
	"	movaps	%%xmm0,	(%%esi)\n"
	"	movaps	%%xmm2,	(%%edi)\n"

	"	add	$8,	%%ebx\n"
	"	add	$16,	%%esi\n"
	"	add	$16,	%%edi\n"

	"	decl	%%eax\n"
	"	jnz	.copy_ij_j_loop\n"

	"	emms\n"
	:"=m"(a), "=m"(b)
	:"m"(c), "m"(fft_size)
	:"memory", "eax", "ebx", "esi", "edi"
	);

}


/*
	this function replaces the following code:
      		for (j=0;j<fft_size;k++) {
         		 c[4*i  ][j].re = real[4*j  ];
         		 c[4*i  ][j].im = imag[4*j  ];
         		 c[4*i+1][j].re = real[4*j+1];
         		 c[4*i+1][j].im = imag[4*j+1];
         		 c[4*i+2][j].re = real[4*j+2];
         		 c[4*i+2][j].im = imag[4*j+2];
         		 c[4*i+3][j].re = real[4*j+3];
         		 c[4*i+3][j].im = imag[4*j+3];
      		}
*/

void float_copy_j_ij(float *a, float *b, complex c[], int fft_size)
{
	__asm __volatile(
	"	movl	%3,	%%eax\n"
	"	movl	%0,	%%esi\n"
	"	movl	%1,	%%edi\n"
	"	movl	%2,	%%ebx\n"

	".balign 16\n"
	".copy_j_ij_loop:\n"

	"	movaps	(%%esi),	%%xmm0\n" // xmm0 = re[0] re[1] re[2] re[3]
	"	movaps	(%%edi),	%%xmm1\n" // xmm1 = im[0] im[1] im[2] im[3]
	"	movaps	%%xmm0,	%%xmm2\n"	  // xmm2 = re[0] re[1] re[2] re[3]
	"	movaps	%%xmm1,	%%xmm3\n"	  // xmm3 = im[0] im[1] im[2] im[3]

	"	unpcklps	%%xmm1,	%%xmm0\n" // xmm0 = re[0] im[0] re[1] im[1]
	"	unpckhps	%%xmm3,	%%xmm2\n" // xmm2 = re[2] im[2] re[3] im[3]

	"	movlps	%%xmm0,	(%%ebx)\n"
	"	movhps	%%xmm0,	 8192(%%ebx)\n"
	"	movlps	%%xmm2,	16384(%%ebx)\n"
	"	movhps	%%xmm2,	24576(%%ebx)\n"

	"	add	$8,	%%ebx\n"
	"	add	$16,	%%esi\n"
	"	add	$16,	%%edi\n"

	"	decl	%%eax\n"
	"	jnz	.copy_j_ij_loop\n"

	"	emms\n"
	:"=m"(a), "=m"(b)
	:"m"(c), "m"(fft_size)
	:"memory", "eax", "ebx", "esi", "edi"
	);

}
#endif //__SHADER_ASM__
